# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from unittest import mock

from neutron_lib.callbacks import events
from neutron_lib.callbacks import exceptions
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants
from neutron_lib import context
from neutron_lib.objects import exceptions as obj_exc
import sqlalchemy
import testtools

from neutron.db import securitygroups_db
from neutron.extensions import security_groups_default_rules as \
    ext_sg_default_rules
from neutron.extensions import securitygroup
from neutron import quota
from neutron.services.revisions import revision_plugin
from neutron.tests.unit import testlib_api


FAKE_SECGROUP = {
    'security_group': {
        "tenant_id": 'fake',
        'description': 'fake',
        'name': 'fake'
    }
}

FAKE_SECGROUP_RULE = {
    'security_group_rule': {
        "tenant_id": 'fake',
        'description': 'fake',
        'name': 'fake',
        'port_range_min': '21',
        'protocol': 'tcp',
        'port_range_max': '23',
        'remote_ip_prefix': '10.0.0.1',
        'ethertype': 'IPv4',
        'remote_group_id': None,
        'remote_address_group_id': None,
        'security_group_id': 'None',
        'direction': 'ingress'
    }
}

DB_PLUGIN_KLASS = 'neutron.db.db_base_plugin_v2.NeutronDbPluginV2'


def fake_callback(resource, event, *args, **kwargs):
    raise KeyError('bar')


class SecurityGroupDbMixinImpl(securitygroups_db.SecurityGroupDbMixin):
    pass


class SecurityGroupDbMixinTestCase(testlib_api.SqlTestCase):

    def setUp(self):
        super(SecurityGroupDbMixinTestCase, self).setUp()
        self.setup_coreplugin(core_plugin=DB_PLUGIN_KLASS)
        self.ctx = context.get_admin_context()
        self.mixin = SecurityGroupDbMixinImpl()
        quota_check = mock.patch.object(quota.QuotaEngine, 'quota_limit_check')
        self.mock_quota_check = quota_check.start()
        is_ext_supported = mock.patch(
            'neutron_lib.api.extensions.is_extension_supported')
        self.is_ext_supported = is_ext_supported.start()
        self.is_ext_supported.return_value = True

    def test_create_security_group_conflict(self):
        with mock.patch.object(registry, "publish") as mock_publish:
            mock_publish.side_effect = exceptions.CallbackFailure(Exception())
            secgroup = {'security_group': mock.ANY}
            with testtools.ExpectedException(
                    securitygroup.SecurityGroupConflict):
                self.mixin.create_security_group(self.ctx, secgroup)

    def test_delete_security_group_in_use(self):
        with mock.patch.object(self.mixin,
                               '_get_port_security_group_bindings'),\
                mock.patch.object(self.mixin, '_get_security_group'),\
                mock.patch.object(registry, "publish") as mock_publish:
            mock_publish.side_effect = exceptions.CallbackFailure(Exception())
            with testtools.ExpectedException(securitygroup.SecurityGroupInUse):
                self.mixin.delete_security_group(self.ctx, mock.ANY)

    def test_update_security_group_statefulness_binded_conflict(self):
        FAKE_SECGROUP['security_group']['stateful'] = mock.ANY
        sg_dict = self.mixin.create_security_group(self.ctx, FAKE_SECGROUP)
        FAKE_SECGROUP['security_group']['stateful'] = not sg_dict['stateful']
        with mock.patch.object(self.mixin,
                               '_get_port_security_group_bindings'), \
                mock.patch.object(registry, "publish") as mock_publish:
            mock_publish.side_effect = exceptions.CallbackFailure(Exception())
            with testtools.ExpectedException(securitygroup.SecurityGroupInUse):
                self.mixin.update_security_group(self.ctx, sg_dict['id'],
                                                 FAKE_SECGROUP)

    def test_update_security_group_conflict(self):
        with mock.patch.object(registry, "publish") as mock_publish:
            mock_publish.side_effect = exceptions.CallbackFailure(Exception())
            secgroup = {'security_group': FAKE_SECGROUP}
            with testtools.ExpectedException(
                    securitygroup.SecurityGroupConflict):
                self.mixin.update_security_group(self.ctx, 'foo_id', secgroup)

    def test_create_security_group_rule_conflict(self):
        with mock.patch.object(self.mixin, '_validate_security_group_rule'),\
                mock.patch.object(self.mixin,
                                  '_check_for_duplicate_rules'),\
                mock.patch.object(registry, "publish") as mock_publish:
            mock_publish.side_effect = exceptions.CallbackFailure(Exception())
            with testtools.ExpectedException(exceptions.CallbackFailure):
                self.mixin.create_security_group_rule(
                    self.ctx, FAKE_SECGROUP_RULE)

    def test__check_for_duplicate_rules_does_not_drop_protocol(self):
        with mock.patch.object(self.mixin, 'get_security_group',
                               return_value=None):
            context = mock.Mock()
            rule_dict = {
                'security_group_rule': {'protocol': None,
                                        'tenant_id': 'fake',
                                        'security_group_id': 'fake',
                                        'direction': 'fake'}
            }
            self.mixin._check_for_duplicate_rules(context, 'fake', [rule_dict])
        self.assertIn('protocol', rule_dict['security_group_rule'])

    def test__check_for_duplicate_rules_ignores_rule_id(self):
        rules = [{'security_group_rule': {'protocol': 'tcp', 'id': 'fake1'}},
                 {'security_group_rule': {'protocol': 'tcp', 'id': 'fake2'}}]

        # NOTE(arosen): the name of this exception is a little misleading
        # in this case as this test, tests that the id fields are dropped
        # while being compared. This is in the case if a plugin specifies
        # the rule ids themselves.
        with mock.patch.object(self.mixin, 'get_security_group',
                               return_value=None):
            self.assertRaises(securitygroup.DuplicateSecurityGroupRuleInPost,
                              self.mixin._check_for_duplicate_rules,
                              context, 'fake', rules)

    def test_check_for_duplicate_diff_rules_remote_ip_prefix_ipv4(self):
        fake_secgroup = copy.deepcopy(FAKE_SECGROUP)
        fake_secgroup['security_group_rules'] = \
            [{'id': 'fake', 'tenant_id': 'fake', 'ethertype': 'IPv4',
              'direction': 'ingress', 'security_group_id': 'fake',
              'remote_ip_prefix': None}]
        with mock.patch.object(self.mixin, 'get_security_group',
                               return_value=fake_secgroup):
            context = mock.Mock()
            rule_dict = {
                'security_group_rule': {'id': 'fake2',
                                        'tenant_id': 'fake',
                                        'security_group_id': 'fake',
                                        'ethertype': 'IPv4',
                                        'direction': 'ingress',
                                        'remote_ip_prefix': '0.0.0.0/0'}
            }
            self.assertRaises(securitygroup.SecurityGroupRuleExists,
                self.mixin._check_for_duplicate_rules,
                context, 'fake', [rule_dict])

    def test_check_for_duplicate_diff_rules_remote_ip_prefix_ipv6(self):
        fake_secgroup = copy.deepcopy(FAKE_SECGROUP)
        fake_secgroup['security_group_rules'] = \
            [{'id': 'fake', 'tenant_id': 'fake', 'ethertype': 'IPv6',
              'direction': 'ingress', 'security_group_id': 'fake',
              'remote_ip_prefix': None}]
        with mock.patch.object(self.mixin, 'get_security_group',
                               return_value=fake_secgroup):
            context = mock.Mock()
            rule_dict = {
                'security_group_rule': {'id': 'fake2',
                                        'tenant_id': 'fake',
                                        'security_group_id': 'fake',
                                        'ethertype': 'IPv6',
                                        'direction': 'ingress',
                                        'remote_ip_prefix': '::/0'}
            }
            self.assertRaises(securitygroup.SecurityGroupRuleExists,
                self.mixin._check_for_duplicate_rules,
                context, 'fake', [rule_dict])

    def test_delete_security_group_rule_in_use(self):
        with mock.patch.object(registry, "publish") as mock_publish:
            mock_publish.side_effect = exceptions.CallbackFailure(Exception())
            with testtools.ExpectedException(exceptions.CallbackFailure):
                self.mixin.delete_security_group_rule(self.ctx, mock.ANY)

    def test_delete_security_group_rule_raise_error_on_not_found(self):
        with testtools.ExpectedException(
                securitygroup.SecurityGroupRuleNotFound):
            self.mixin.delete_security_group_rule(self.ctx, 'foo_rule')

    def test_validate_ethertype_and_protocol(self):
        fake_ipv4_rules = [{'protocol': constants.PROTO_NAME_IPV6_ICMP,
                            'ethertype': constants.IPv4},
                           {'protocol': constants.PROTO_NAME_IPV6_ICMP_LEGACY,
                            'ethertype': constants.IPv4},
                           {'protocol': constants.PROTO_NAME_IPV6_ENCAP,
                            'ethertype': constants.IPv4},
                           {'protocol': constants.PROTO_NAME_IPV6_ROUTE,
                            'ethertype': constants.IPv4},
                           {'protocol': constants.PROTO_NAME_IPV6_FRAG,
                            'ethertype': constants.IPv4},
                           {'protocol': constants.PROTO_NAME_IPV6_NONXT,
                            'ethertype': constants.IPv4},
                           {'protocol': constants.PROTO_NAME_IPV6_OPTS,
                            'ethertype': constants.IPv4},
                           {'protocol': str(constants.PROTO_NUM_IPV6_ICMP),
                            'ethertype': constants.IPv4},
                           {'protocol': str(constants.PROTO_NUM_IPV6_ENCAP),
                            'ethertype': constants.IPv4},
                           {'protocol': str(constants.PROTO_NUM_IPV6_ROUTE),
                            'ethertype': constants.IPv4},
                           {'protocol': str(constants.PROTO_NUM_IPV6_FRAG),
                            'ethertype': constants.IPv4},
                           {'protocol': str(constants.PROTO_NUM_IPV6_NONXT),
                            'ethertype': constants.IPv4},
                           {'protocol': str(constants.PROTO_NUM_IPV6_OPTS),
                            'ethertype': constants.IPv4}]
        # test wrong protocols
        for rule in fake_ipv4_rules:
            with testtools.ExpectedException(
                    securitygroup.SecurityGroupEthertypeConflictWithProtocol):
                self.mixin._validate_ethertype_and_protocol(rule)

    def test_security_group_precommit_create_event_fail(self):
        registry.subscribe(fake_callback, resources.SECURITY_GROUP,
                           events.PRECOMMIT_CREATE)
        with mock.patch.object(sqlalchemy.orm.session.SessionTransaction,
                               'rollback') as mock_rollback:
            self.assertRaises(securitygroup.SecurityGroupConflict,
                              self.mixin.create_security_group,
                              self.ctx, FAKE_SECGROUP)
            self.assertTrue(mock_rollback.called)

    def test_security_group_precommit_update_event_fail(self):
        registry.subscribe(fake_callback, resources.SECURITY_GROUP,
                           events.PRECOMMIT_UPDATE)
        sg_dict = self.mixin.create_security_group(self.ctx, FAKE_SECGROUP)
        with mock.patch.object(sqlalchemy.orm.session.SessionTransaction,
                              'rollback') as mock_rollback:
            self.assertRaises(securitygroup.SecurityGroupConflict,
                              self.mixin.update_security_group,
                              self.ctx, sg_dict['id'], FAKE_SECGROUP)
            self.assertTrue(mock_rollback.called)

    def test_security_group_precommit_delete_event_fail(self):
        registry.subscribe(fake_callback, resources.SECURITY_GROUP,
                           events.PRECOMMIT_DELETE)
        sg_dict = self.mixin.create_security_group(self.ctx, FAKE_SECGROUP)
        with mock.patch.object(sqlalchemy.orm.session.SessionTransaction,
                              'rollback') as mock_rollback:
            self.assertRaises(securitygroup.SecurityGroupInUse,
                              self.mixin.delete_security_group,
                              self.ctx, sg_dict['id'])
            self.assertTrue(mock_rollback.called)

    def _test_security_group_precommit_create_event(self,
                                                    with_revisions=False):
        DEFAULT_SECGROUP = {
            'tenant_id': FAKE_SECGROUP['security_group']['tenant_id'],
            'name': 'default',
            'description': 'Default security group',
        }
        DEFAULT_SECGROUP_DICT = {
            'id': mock.ANY,
            'tenant_id': FAKE_SECGROUP['security_group']['tenant_id'],
            'project_id': FAKE_SECGROUP['security_group']['tenant_id'],
            'name': 'default',
            'description': 'Default security group',
            'stateful': mock.ANY,
            'standard_attr_id': mock.ANY,
            'shared': False,
            'security_group_rules': [
                # 2 Custom rules from template
                mock.ANY, mock.ANY
            ],
        }
        if with_revisions:
            DEFAULT_SECGROUP_DICT.update({
                'revision_number': mock.ANY,
            })
        with mock.patch.object(registry, 'publish') as publish, \
                mock.patch.object(
                    self.mixin, 'get_default_security_group_rules',
                    return_value=[mock.MagicMock(), mock.MagicMock()]):
            sg_dict = self.mixin.create_security_group(self.ctx, FAKE_SECGROUP)

            publish.assert_has_calls([
                mock.call(resources.SECURITY_GROUP, 'before_create', mock.ANY,
                          payload=mock.ANY),
                mock.call(resources.SECURITY_GROUP, 'before_create', mock.ANY,
                          payload=mock.ANY),
                mock.call(resources.SECURITY_GROUP, 'precommit_create',
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.SECURITY_GROUP, 'after_create', mock.ANY,
                          payload=mock.ANY),
                mock.call(resources.SECURITY_GROUP, 'precommit_create',
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.SECURITY_GROUP, 'after_create', mock.ANY,
                          payload=mock.ANY)])

            payload = publish.mock_calls[0][2]['payload']
            self.assertDictEqual(payload.desired_state,
                                 FAKE_SECGROUP['security_group'])

            payload = publish.mock_calls[1][2]['payload']
            self.assertDictEqual(payload.desired_state,
                                 DEFAULT_SECGROUP)

            payload = publish.mock_calls[2][2]['payload']
            self.assertDictEqual(payload.latest_state,
                                 DEFAULT_SECGROUP_DICT)
            self.assertTrue(payload.metadata['is_default'])

            payload = publish.mock_calls[3][2]['payload']
            self.assertDictEqual(payload.latest_state,
                                 DEFAULT_SECGROUP_DICT)
            self.assertTrue(payload.metadata['is_default'])

            payload = publish.mock_calls[4][2]['payload']
            self.assertDictEqual(payload.latest_state, sg_dict)
            self.assertFalse(payload.metadata['is_default'])

            payload = publish.mock_calls[5][2]['payload']
            self.assertDictEqual(payload.latest_state, sg_dict)
            self.assertFalse(payload.metadata['is_default'])

            # Ensure that the result of create is same as get.
            # Especially we want to check the revision number here.
            sg_dict_got = self.mixin.get_security_group(
                self.ctx, sg_dict['id'])
            self.assertEqual(sg_dict, sg_dict_got)

    def test_security_group_precommit_create_event_with_revisions(self):
        revision = revision_plugin.RevisionPlugin()
        self._test_security_group_precommit_create_event(True)
        del revision  # appease pep8

    def test_security_group_precommit_create_event(self):
        self._test_security_group_precommit_create_event()

    def test_security_group_precommit_update_event(self):
        FAKE_SECGROUP['security_group']['stateful'] = mock.ANY
        original_sg_dict = self.mixin.create_security_group(self.ctx,
                                                            FAKE_SECGROUP)
        sg_id = original_sg_dict['id']
        with mock.patch.object(self.mixin,
                               '_get_port_security_group_bindings'), \
                mock.patch.object(registry, "publish") as mock_publish:
            fake_secgroup = copy.deepcopy(FAKE_SECGROUP)
            fake_secgroup['security_group']['name'] = 'updated_fake'
            fake_secgroup['security_group']['stateful'] = mock.ANY
            sg_dict = self.mixin.update_security_group(
                    self.ctx, sg_id, fake_secgroup)

            mock_publish.assert_has_calls(
                [mock.call(resources.SECURITY_GROUP, events.PRECOMMIT_UPDATE,
                           mock.ANY, payload=mock.ANY)])
            payload = mock_publish.call_args[1]['payload']
            self.assertEqual(original_sg_dict, payload.states[0])
            self.assertEqual(sg_id, payload.resource_id)
            self.assertEqual(sg_dict, payload.latest_state)

    def test_security_group_precommit_and_after_delete_event(self):
        with mock.patch.object(
                self.mixin, 'get_default_security_group_rules',
                return_value=[mock.MagicMock(), mock.MagicMock()]):
            sg_dict = self.mixin.create_security_group(self.ctx, FAKE_SECGROUP)
        with mock.patch.object(registry, "publish") as mock_publish:
            self.mixin.delete_security_group(self.ctx, sg_dict['id'])
            sg_dict['security_group_rules'] = mock.ANY
            mock_publish.assert_has_calls(
                [mock.call('security_group', 'precommit_delete',
                           mock.ANY, payload=mock.ANY),
                 mock.call('security_group', 'after_delete',
                           mock.ANY,
                           payload=mock.ANY)])
            payload = mock_publish.mock_calls[1][2]['payload']
            self.assertEqual(mock.ANY, payload.context)
            self.assertEqual(sg_dict, payload.latest_state)
            self.assertEqual(sg_dict['id'], payload.resource_id)
            self.assertEqual([mock.ANY, mock.ANY],
                payload.metadata.get('security_group_rule_ids'))

            payload = mock_publish.mock_calls[2][2]['payload']
            self.assertEqual(mock.ANY, payload.context)
            self.assertEqual(sg_dict, payload.latest_state)
            self.assertEqual(sg_dict['id'], payload.resource_id)
            self.assertEqual([mock.ANY, mock.ANY],
                             payload.metadata.get('security_group_rule_ids'))

    def test_security_group_rule_precommit_create_event_fail(self):
        registry.subscribe(fake_callback, resources.SECURITY_GROUP_RULE,
                           events.PRECOMMIT_CREATE)
        sg_dict = self.mixin.create_security_group(self.ctx, FAKE_SECGROUP)
        fake_rule = FAKE_SECGROUP_RULE
        fake_rule['security_group_rule']['security_group_id'] = sg_dict['id']
        with mock.patch.object(sqlalchemy.orm.session.SessionTransaction,
                               'rollback') as mock_rollback,\
                mock.patch.object(self.mixin, '_get_security_group'):
            self.assertRaises(securitygroup.SecurityGroupConflict,
                              self.mixin.create_security_group_rule,
                              self.ctx, fake_rule)
            self.assertTrue(mock_rollback.called)

    def test_security_group_rule_precommit_delete_event_fail(self):
        registry.subscribe(fake_callback, resources.SECURITY_GROUP_RULE,
                           events.PRECOMMIT_DELETE)
        sg_dict = self.mixin.create_security_group(self.ctx, FAKE_SECGROUP)
        fake_rule = FAKE_SECGROUP_RULE
        fake_rule['security_group_rule']['security_group_id'] = sg_dict['id']
        with mock.patch.object(sqlalchemy.orm.session.SessionTransaction,
                               'rollback') as mock_rollback,\
                mock.patch.object(self.mixin, '_get_security_group'):
            sg_rule_dict = self.mixin.create_security_group_rule(self.ctx,
                   fake_rule)
            self.assertRaises(securitygroup.SecurityGroupRuleInUse,
                              self.mixin.delete_security_group_rule, self.ctx,
                              sg_rule_dict['id'])
            self.assertTrue(mock_rollback.called)

    def test_security_group_rule_precommit_create_event(self):
        sg_dict = self.mixin.create_security_group(self.ctx, FAKE_SECGROUP)
        fake_rule = FAKE_SECGROUP_RULE
        fake_rule['security_group_rule']['security_group_id'] = sg_dict['id']
        with mock.patch.object(registry, "publish") as mock_publish, \
                mock.patch.object(self.mixin, '_get_security_group'):
            sg_rule = self.mixin.create_security_group_rule(self.ctx,
                                                            fake_rule)
            mock_publish.assert_has_calls([mock.call(
                'security_group_rule',
                'precommit_create', mock.ANY, payload=mock.ANY)])

            payload = mock_publish.mock_calls[1][2]['payload']
            self.assertEqual(sg_rule['id'], payload.resource_id)
            self.assertEqual(sg_rule, payload.latest_state)

    def test_sg_rule_before_precommit_and_after_delete_event(self):
        sg_dict = self.mixin.create_security_group(self.ctx, FAKE_SECGROUP)
        fake_rule = FAKE_SECGROUP_RULE
        fake_rule['security_group_rule']['security_group_id'] = sg_dict['id']
        with mock.patch.object(registry, "publish") as mock_publish, \
                mock.patch.object(self.mixin, '_get_security_group'):
            sg_rule_dict = self.mixin.create_security_group_rule(self.ctx,
                                                                 fake_rule)
            self.mixin.delete_security_group_rule(self.ctx,
                                                  sg_rule_dict['id'])
            mock_publish.assert_has_calls([mock.call('security_group_rule',
                                                     'before_delete',
                                                     mock.ANY,
                                                     payload=mock.ANY)])
            mock_publish.assert_has_calls([mock.call('security_group_rule',
                                                    'precommit_delete',
                                                     mock.ANY,
                                                     payload=mock.ANY)])
            mock_publish.assert_has_calls([mock.call('security_group_rule',
                                                    'after_delete',
                                                     mock.ANY,
                                                     payload=mock.ANY)])

            payload = mock_publish.mock_calls[1][2]['payload']
            self.assertEqual(mock.ANY, payload.context)
            self.assertEqual(sg_rule_dict, payload.latest_state)
            self.assertEqual(sg_rule_dict['id'], payload.resource_id)

            payload = mock_publish.mock_calls[2][2]['payload']
            self.assertEqual(mock.ANY, payload.context)
            self.assertEqual(sg_rule_dict, payload.latest_state)
            self.assertEqual(sg_rule_dict['id'], payload.resource_id)

            payload = mock_publish.mock_calls[3][2]['payload']
            self.assertEqual(mock.ANY, payload.context)
            self.assertEqual(sg_rule_dict['id'], payload.resource_id)

    def test_get_ip_proto_name_and_num(self):
        protocols = [constants.PROTO_NAME_UDP, str(constants.PROTO_NUM_TCP),
                     'blah', '111']
        protocol_names_nums = (
            [[constants.PROTO_NAME_UDP, str(constants.PROTO_NUM_UDP)],
             [constants.PROTO_NAME_TCP, str(constants.PROTO_NUM_TCP)],
             ['blah', 'blah'], ['111', '111']])

        for i, protocol in enumerate(protocols):
            self.assertEqual(protocol_names_nums[i],
                             self.mixin._get_ip_proto_name_and_num(protocol))

    def test__validate_port_range_for_icmp_exception(self):
        states = [(1, 256, securitygroup.SecurityGroupInvalidIcmpValue),
                  (None, 6, securitygroup.SecurityGroupMissingIcmpType),
                  (300, 1, securitygroup.SecurityGroupInvalidIcmpValue)]
        for protocol in (constants.PROTO_NAME_ICMP,
                         constants.PROTO_NAME_IPV6_ICMP,
                         constants.PROTO_NAME_IPV6_ICMP_LEGACY):
            for pmin, pmax, exception in states:
                self.assertRaises(exception,
                    self.mixin._validate_port_range,
                    {'port_range_min': pmin,
                     'port_range_max': pmax,
                     'protocol': protocol})

    def test__validate_port_range_exception(self):
        self.assertRaises(securitygroup.SecurityGroupInvalidPortValue,
                          self.mixin._validate_port_range,
                          {'port_range_min': 0,
                           'port_range_max': None,
                           'protocol': constants.PROTO_NAME_TCP})
        self.assertRaises(securitygroup.SecurityGroupInvalidPortRange,
                          self.mixin._validate_port_range,
                          {'port_range_min': 1,
                           'port_range_max': None,
                           'protocol': constants.PROTO_NAME_SCTP})
        self.assertRaises(securitygroup.SecurityGroupInvalidPortRange,
                          self.mixin._validate_port_range,
                          {'port_range_min': 1000,
                           'port_range_max': 1,
                           'protocol': constants.PROTO_NAME_UDPLITE})
        self.assertRaises(
            securitygroup.SecurityGroupInvalidProtocolForPort,
            self.mixin._validate_port_range,
            {'port_range_min': 100,
             'port_range_max': 200,
             'protocol': '111'})
        self.assertRaises(
            securitygroup.SecurityGroupInvalidProtocolForPort,
            self.mixin._validate_port_range,
            {'port_range_min': 100,
             'port_range_max': None,
             'protocol': constants.PROTO_NAME_VRRP})
        self.assertRaises(
            securitygroup.SecurityGroupInvalidProtocolForPort,
            self.mixin._validate_port_range,
            {'port_range_min': None,
             'port_range_max': 200,
             'protocol': constants.PROTO_NAME_VRRP})

    def _create_environment(self):
        self.sg = copy.deepcopy(FAKE_SECGROUP)
        self.user_ctx = context.Context(user_id='user1', tenant_id='tenant_1',
                                        is_admin=False, overwrite=False)
        self.admin_ctx = context.Context(user_id='user2', tenant_id='tenant_2',
                                         is_admin=True, overwrite=False)
        self.sg_user = self.mixin.create_security_group(
            self.user_ctx, {'security_group': {'name': 'name',
                                               'tenant_id': 'tenant_1',
                                               'description': 'fake'}})

    def test_get_security_group_rules(self):
        self._create_environment()
        rules_before = self.mixin.get_security_group_rules(self.user_ctx)

        rule = copy.deepcopy(FAKE_SECGROUP_RULE)
        rule['security_group_rule']['security_group_id'] = self.sg_user['id']
        rule['security_group_rule']['tenant_id'] = 'tenant_2'
        self.mixin.create_security_group_rule(self.admin_ctx, rule)

        rules_after = self.mixin.get_security_group_rules(self.user_ctx)
        self.assertEqual(len(rules_before) + 1, len(rules_after))
        for rule in (rule for rule in rules_after if rule not in rules_before):
            self.assertEqual('tenant_2', rule['tenant_id'])

    def test_get_security_group_rules_filters_passed(self):
        self._create_environment()
        filters = {'security_group_id': self.sg_user['id']}
        rules_before = self.mixin.get_security_group_rules(self.user_ctx,
                                                           filters=filters)

        default_sg = self.mixin.get_security_groups(
            self.user_ctx, filters={'name': 'default'})[0]
        rule = copy.deepcopy(FAKE_SECGROUP_RULE)
        rule['security_group_rule']['security_group_id'] = default_sg['id']
        rule['security_group_rule']['tenant_id'] = 'tenant_1'
        self.mixin.create_security_group_rule(self.user_ctx, rule)

        rules_after = self.mixin.get_security_group_rules(self.user_ctx,
                                                          filters=filters)
        self.assertEqual(rules_before, rules_after)

    def test_get_security_group_rules_admin_context(self):
        self._create_environment()
        rules_before = self.mixin.get_security_group_rules(self.ctx)

        rule = copy.deepcopy(FAKE_SECGROUP_RULE)
        rule['security_group_rule']['security_group_id'] = self.sg_user['id']
        rule['security_group_rule']['tenant_id'] = 'tenant_1'
        self.mixin.create_security_group_rule(self.user_ctx, rule)

        rules_after = self.mixin.get_security_group_rules(self.ctx)
        self.assertEqual(len(rules_before) + 1, len(rules_after))
        for rule in (rule for rule in rules_after if rule not in rules_before):
            self.assertEqual('tenant_1', rule['tenant_id'])
            self.assertEqual(self.sg_user['id'], rule['security_group_id'])

    def test__ensure_default_security_group(self):
        with mock.patch.object(
                self.mixin, '_get_default_sg_id') as get_default_sg_id,\
                mock.patch.object(
                        self.mixin, 'create_security_group') as create_sg:
            get_default_sg_id.return_value = None
            self.mixin._ensure_default_security_group(self.ctx, 'tenant_1')
            create_sg.assert_called_once_with(
                self.ctx,
                {'security_group': {
                    'name': 'default',
                    'tenant_id': 'tenant_1',
                    'description': securitygroups_db.DEFAULT_SG_DESCRIPTION}},
                default_sg=True)
            get_default_sg_id.assert_called_once_with(self.ctx, 'tenant_1')

    def test__ensure_default_security_group_already_exists(self):
        with mock.patch.object(
                self.mixin, '_get_default_sg_id') as get_default_sg_id,\
                mock.patch.object(
                        self.mixin, 'create_security_group') as create_sg:
            get_default_sg_id.return_value = 'default_sg_id'
            self.mixin._ensure_default_security_group(self.ctx, 'tenant_1')
            create_sg.assert_not_called()
            get_default_sg_id.assert_called_once_with(self.ctx, 'tenant_1')

    def test__ensure_default_security_group_created_in_parallel(self):
        with mock.patch.object(
                self.mixin, '_get_default_sg_id') as get_default_sg_id,\
                mock.patch.object(
                        self.mixin, 'create_security_group') as create_sg:
            get_default_sg_id.side_effect = [None, 'default_sg_id']
            create_sg.side_effect = obj_exc.NeutronDbObjectDuplicateEntry(
                mock.Mock(), mock.Mock())
            self.mixin._ensure_default_security_group(self.ctx, 'tenant_1')
            create_sg.assert_called_once_with(
                self.ctx,
                {'security_group': {
                    'name': 'default',
                    'tenant_id': 'tenant_1',
                    'description': securitygroups_db.DEFAULT_SG_DESCRIPTION}},
                default_sg=True)
            get_default_sg_id.assert_has_calls([
                mock.call(self.ctx, 'tenant_1'),
                mock.call(self.ctx, 'tenant_1')])

    def test__ensure_default_security_group_when_disabled(self):
        with mock.patch.object(
                    self.mixin, '_get_default_sg_id') as get_default_sg_id,\
                mock.patch.object(
                        self.mixin, 'create_security_group') as create_sg:
            self.is_ext_supported.return_value = False
            self.mixin._ensure_default_security_group(self.ctx, 'tenant_1')
            create_sg.assert_not_called()
            get_default_sg_id.assert_not_called()

    def test__ensure_default_security_group_tenant_mismatch(self):
        with mock.patch.object(
                self.mixin, '_get_default_sg_id') as get_default_sg_id,\
                mock.patch.object(
                        self.mixin, 'create_security_group') as create_sg:
            context = mock.Mock()
            context.tenant_id = 'tenant_0'
            context.is_admin = False
            self.mixin._ensure_default_security_group(context, 'tenant_1')
            create_sg.assert_not_called()
            get_default_sg_id.assert_not_called()

    def test__check_for_duplicate_default_rules_does_not_drop_protocol(self):
        with mock.patch.object(self.mixin, 'get_default_security_group_rules',
                               return_value=None):
            context = mock.Mock()
            rule_dict = {
                'default_security_group_rule': {'protocol': None,
                                                'direction': 'fake'}
            }
            self.mixin._check_for_duplicate_default_rules(
                context, [rule_dict])
        self.assertIn('protocol', rule_dict['default_security_group_rule'])

    def test__check_for_duplicate_default_rules_ignores_rule_id(self):
        rules = [
            {'default_security_group_rule': {
                'protocol': 'tcp', 'id': 'fake1'}},
            {'default_security_group_rule': {
                'protocol': 'tcp', 'id': 'fake2'}}]

        # NOTE(arosen): the name of this exception is a little misleading
        # in this case as this test, tests that the id fields are dropped
        # while being compared. This is in the case if a plugin specifies
        # the rule ids themselves.
        with mock.patch.object(self.mixin, 'get_default_security_group_rules',
                               return_value=None):
            self.assertRaises(
                ext_sg_default_rules.DuplicateDefaultSgRuleInPost,
                self.mixin._check_for_duplicate_default_rules, context, rules)

    def test__check_for_duplicate_default_rules_rule_used_in_non_default_sg(
            self):
        fake_rules = [
            {'id': 'fake',
             'used_in_default_sg': True,
             'used_in_non_default_sg': True}
        ]
        with mock.patch.object(self.mixin, 'get_default_security_group_rules',
                               return_value=fake_rules):
            context = mock.Mock()
            rule_dict = {
                'default_security_group_rule': {
                    'id': 'fake2',
                    'used_in_default_sg': False,
                    'used_in_non_default_sg': True}
            }
            self.assertRaises(
                ext_sg_default_rules.DefaultSecurityGroupRuleExists,
                self.mixin._check_for_duplicate_default_rules,
                context, [rule_dict])

    def test__check_for_duplicate_default_rules_rule_used_in_default_sg(
            self):
        fake_rules = [
            {'id': 'fake',
             'used_in_default_sg': True,
             'used_in_non_default_sg': True}
        ]
        with mock.patch.object(self.mixin, 'get_default_security_group_rules',
                               return_value=fake_rules):
            context = mock.Mock()
            rule_dict = {
                'default_security_group_rule': {
                    'id': 'fake2',
                    'used_in_default_sg': True,
                    'used_in_non_default_sg': False}
            }
            self.assertRaises(
                ext_sg_default_rules.DefaultSecurityGroupRuleExists,
                self.mixin._check_for_duplicate_default_rules,
                context, [rule_dict])

    def test__check_for_duplicate_diff_default_rules_remote_ip_prefix_ipv4(
            self):
        fake_rules = [
            {'id': 'fake', 'ethertype': 'IPv4',
             'direction': 'ingress', 'remote_ip_prefix': None}
        ]
        with mock.patch.object(self.mixin, 'get_default_security_group_rules',
                               return_value=fake_rules):
            context = mock.Mock()
            rule_dict = {
                'default_security_group_rule': {
                    'id': 'fake2',
                    'ethertype': 'IPv4',
                    'direction': 'ingress',
                    'remote_ip_prefix': '0.0.0.0/0'}
            }
            self.assertRaises(
                ext_sg_default_rules.DefaultSecurityGroupRuleExists,
                self.mixin._check_for_duplicate_default_rules,
                context, [rule_dict])

    def test__check_for_duplicate_diff_default_rules_remote_ip_prefix_ipv6(
            self):
        fake_rules = [
            {'id': 'fake', 'ethertype': 'IPv6',
             'direction': 'ingress', 'remote_ip_prefix': None}
        ]
        with mock.patch.object(self.mixin, 'get_default_security_group_rules',
                               return_value=fake_rules):
            context = mock.Mock()
            rule_dict = {
                'default_security_group_rule': {
                    'id': 'fake2',
                    'ethertype': 'IPv6',
                    'direction': 'ingress',
                    'remote_ip_prefix': '::/0'}
            }
            self.assertRaises(
                ext_sg_default_rules.DefaultSecurityGroupRuleExists,
                self.mixin._check_for_duplicate_default_rules,
                context, [rule_dict])



# Copyright (c) 2018 Fujitsu Limited
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.services.logapi import constants as log_const
from oslo_log import log as logging

from neutron.objects import router
from neutron.services.logapi.common import exceptions as log_exc
from neutron.services.logapi.common import validators

LOG = logging.getLogger(__name__)

EVENTS_DISABLE = [log_const.DROP_EVENT, log_const.ACCEPT_EVENT]


def _get_router(context, router_id):
    router_obj = router.Router.get_object(context, id=router_id)
    if not router_obj:
        raise log_exc.ResourceNotFound(resource_id=router_id)
    return router_obj


@validators.ResourceValidateRequest.register(log_const.SNAT)
def validate_snat_request(context, log_data):
    """Validate the incoming SNAT log request

    This method validates whether SNAT log request is satisfied or not.

    A ResourceNotFound will be raised if resource_id in log_data does not
    belong to any Router object. This method will also raise a
    RouterNotEnabledSnat exception in the case of a indicated router does not
    enable SNAT feature.
    """

    resource_id = log_data.get('resource_id')
    event = log_data.get('event')
    if not resource_id:
        raise log_exc.ResourceIdNotSpecified(resource_type=log_const.SNAT)
    if event in EVENTS_DISABLE:
        raise log_exc.EventsDisabled(events=EVENTS_DISABLE,
                                     resource_type=log_const.SNAT)
    router_obj = _get_router(context, resource_id)
    # Check whether SNAT is enabled or not
    if not router_obj.enable_snat:
        raise log_exc.RouterNotEnabledSnat(resource_id=resource_id)
    # Check whether router gateway is set or not.
    if not router_obj.gw_port_id:
        raise log_exc.RouterGatewayNotSet(resource_id=resource_id)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import ddt
import netaddr
from neutron_lib.api.definitions import external_net as extnet_apidef
from neutron_lib.api.definitions import l3 as l3_apidef
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as n_const
from neutron_lib import context
from neutron_lib.db import api as db_api
from neutron_lib import exceptions as n_exc
from neutron_lib.exceptions import extraroute as xroute_exc
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from neutron_lib.plugins import utils as plugin_utils
from oslo_utils import uuidutils
import testtools
import webob.exc

from neutron.db import extraroute_db
from neutron.db import l3_db
from neutron.db.models import l3 as l3_models
from neutron.db.models import l3_attrs
from neutron.db import models_v2
from neutron.extensions import segment as segment_ext
from neutron.objects import base as base_obj
from neutron.objects import network as network_obj
from neutron.objects import ports as port_obj
from neutron.objects import router as l3_obj
from neutron.objects import subnet as subnet_obj
from neutron.tests import base
from neutron.tests.unit.db import test_db_base_plugin_v2


@ddt.ddt
class TestL3_NAT_dbonly_mixin(
        test_db_base_plugin_v2.NeutronDbPluginV2TestCase):

    def setUp(self, *args, **kwargs):
        super(TestL3_NAT_dbonly_mixin, self).setUp(*args, **kwargs)
        # "extraroute_db.ExtraRoute_dbonly_mixin" inherits from
        # "l3_db.L3_NAT_dbonly_mixin()", the class under test. This is used
        # instead to test the validation of router routes and GW change because
        # implements "_validate_routes".
        self.db = extraroute_db.ExtraRoute_dbonly_mixin()
        self.ctx = mock.Mock()

    def test__each_port_having_fixed_ips_none(self):
        """Be sure the method returns an empty list when None is passed"""
        filtered = l3_db.L3_NAT_dbonly_mixin._each_port_having_fixed_ips(None)
        self.assertEqual([], list(filtered))

    def test__new__passes_args(self):
        class T(l3_db.L3_NAT_db_mixin):
            def __init__(self, *args, **kwargs):
                self.args = args
                self.kwargs = kwargs

        t = T(1, 2, a=3)
        self.assertEqual((1, 2), t.args)
        self.assertEqual({'a': 3}, t.kwargs)

    def test__each_port_having_fixed_ips(self):
        """Basic test that ports without fixed ips are filtered out"""
        ports = [{'id': 'a', 'fixed_ips': [mock.sentinel.fixedip]},
                 {'id': 'b'}]
        filtered = l3_db.L3_NAT_dbonly_mixin._each_port_having_fixed_ips(ports)
        ids = [p['id'] for p in filtered]
        self.assertEqual(['a'], ids)

    def test__get_subnets_by_network_no_query(self):
        """Basic test that no query is performed if no Ports are passed"""
        context = mock.Mock()
        with mock.patch.object(directory, 'get_plugin') as get_p:
            self.db._get_subnets_by_network_list(context, [])
        self.assertFalse(context.session.query.called)
        self.assertFalse(get_p.called)

    def test__get_subnets_by_network(self):
        """Basic test that the right query is called"""
        context = mock.MagicMock()
        query = context.session.query().outerjoin().filter()
        query.__iter__.return_value = [(mock.sentinel.subnet_db,
                                        mock.sentinel.address_scope_id)]

        with mock.patch.object(directory, 'get_plugin') as get_p:
            get_p()._make_subnet_dict.return_value = {
                'network_id': mock.sentinel.network_id}
            subnets = self.db._get_subnets_by_network_list(
                context, [mock.sentinel.network_id])
        self.assertEqual({
            mock.sentinel.network_id: [{
                'address_scope_id': mock.sentinel.address_scope_id,
                'network_id': mock.sentinel.network_id}]}, subnets)

    def test__get_mtus_by_network_list(self):
        """Basic test that the query get_networks is correctly"""
        network = {'id': mock.sentinel.network_id,
                   'name': mock.sentinel.name,
                   'mtu': mock.sentinel.mtu}
        with mock.patch.object(directory, 'get_plugin') as get_p:
            get_p().get_networks.return_value = [network]
            result = self.db._get_mtus_by_network_list(
                mock.sentinel.context, [mock.sentinel.network_id])
            get_p().get_networks.assert_called_once_with(
                mock.sentinel.context,
                filters={'id': [mock.sentinel.network_id]},
                fields=['id', 'mtu'])
            self.assertEqual({mock.sentinel.network_id: mock.sentinel.mtu},
                             result)

    def test__populate_ports_for_subnets_none(self):
        """Basic test that the method runs correctly with no ports"""
        ports = []
        with mock.patch.object(directory, 'get_plugin') as get_p:
            get_p().get_networks.return_value = []
            self.db._populate_mtu_and_subnets_for_ports(mock.sentinel.context,
                                                        ports)
        self.assertEqual([], ports)

    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin,
                       '_get_subnets_by_network_list')
    def test__populate_ports_for_subnets(self, get_subnets_by_network):
        cidr = "2001:db8::/64"
        subnet = {'id': mock.sentinel.subnet_id,
                  'cidr': cidr,
                  'gateway_ip': mock.sentinel.gateway_ip,
                  'dns_nameservers': mock.sentinel.dns_nameservers,
                  'ipv6_ra_mode': mock.sentinel.ipv6_ra_mode,
                  'subnetpool_id': mock.sentinel.subnetpool_id,
                  'address_scope_id': mock.sentinel.address_scope_id}
        get_subnets_by_network.return_value = {'net_id': [subnet]}

        ports = [{'network_id': 'net_id',
                  'id': 'port_id',
                  'fixed_ips': [{'subnet_id': mock.sentinel.subnet_id}],
                  'device_owner': 'compute:nova'}]
        with mock.patch.object(directory, 'get_plugin') as get_p:
            get_p().get_networks.return_value = [{'id': 'net_id', 'mtu': 1446}]
            self.db._populate_mtu_and_subnets_for_ports(mock.sentinel.context,
                                                        ports)
            keys = ('id', 'cidr', 'gateway_ip', 'ipv6_ra_mode',
                    'subnetpool_id', 'dns_nameservers')
            address_scopes = {4: None, 6: mock.sentinel.address_scope_id}
            self.assertEqual([{'extra_subnets': [],
                               'fixed_ips': [{'subnet_id':
                                              mock.sentinel.subnet_id,
                                              'prefixlen': 64}],
                               'id': 'port_id',
                               'mtu': 1446,
                               'network_id': 'net_id',
                               'subnets': [{k: subnet[k] for k in keys}],
                               'address_scopes': address_scopes,
                               'device_owner': 'compute:nova'}], ports)

    @ddt.unpack
    @ddt.data({'plugin_loaded': False, 'seg1': None, 'seg2': None},
              {'plugin_loaded': True, 'seg1': None, 'seg2': None},
              {'plugin_loaded': True, 'seg1': 'seg1', 'seg2': 'seg2'})
    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin,
                       '_get_subnets_by_network_list')
    def test__populate_ports_for_subnets_gw_port(self, get_subnets_by_network,
                                                 plugin_loaded, seg1, seg2):
        subnets = [
            {'id': uuidutils.generate_uuid(),
             'cidr': '10.1.0.0/24',
             'gateway_ip': mock.sentinel.gateway_ip,
             'dns_nameservers': mock.sentinel.dns_nameservers,
             'ipv6_ra_mode': mock.sentinel.ipv6_ra_mode,
             'subnetpool_id': mock.sentinel.subnetpool_id,
             'address_scope_id': mock.sentinel.address_scope_id,
             'segment_id': seg1},
            {'id': uuidutils.generate_uuid(),
             'cidr': '10.2.0.0/24',
             'gateway_ip': mock.sentinel.gateway_ip,
             'dns_nameservers': mock.sentinel.dns_nameservers,
             'ipv6_ra_mode': mock.sentinel.ipv6_ra_mode,
             'subnetpool_id': mock.sentinel.subnetpool_id,
             'address_scope_id': mock.sentinel.address_scope_id,
             'segment_id': seg1},
            {'id': uuidutils.generate_uuid(),
             'cidr': '10.3.0.0/24',
             'gateway_ip': mock.sentinel.gateway_ip,
             'dns_nameservers': mock.sentinel.dns_nameservers,
             'ipv6_ra_mode': mock.sentinel.ipv6_ra_mode,
             'subnetpool_id': mock.sentinel.subnetpool_id,
             'address_scope_id': mock.sentinel.address_scope_id,
             'segment_id': seg2}]
        get_subnets_by_network.return_value = {'net_id': subnets}

        ports = [{'network_id': 'net_id',
                  'id': 'port_id',
                  'fixed_ips': [{'subnet_id': subnets[0]['id']}],
                  'device_owner': n_const.DEVICE_OWNER_ROUTER_GW}]
        with mock.patch.object(directory, 'get_plugin') as get_p, \
                mock.patch.object(segment_ext.SegmentPluginBase,
                                  'is_loaded', return_value=plugin_loaded):
            get_p().get_networks.return_value = [{'id': 'net_id', 'mtu': 1446}]
            self.db._populate_mtu_and_subnets_for_ports(mock.sentinel.context,
                                                        ports)
            keys = ('id', 'cidr', 'gateway_ip', 'ipv6_ra_mode',
                    'subnetpool_id', 'dns_nameservers')
            address_scopes = {4: mock.sentinel.address_scope_id, 6: None}
            reference = {'fixed_ips': [{'subnet_id': subnets[0]['id'],
                                        'prefixlen': 24}],
                         'id': 'port_id',
                         'mtu': 1446,
                         'network_id': 'net_id',
                         'subnets': [{k: subnets[0][k] for k in keys}],
                         'address_scopes': address_scopes,
                         'device_owner': n_const.DEVICE_OWNER_ROUTER_GW,
                         'extra_subnets': [{k: subnets[1][k] for k in keys}]}
            # If RPN plugin is not enabled or the network subnets do not have
            # associated segments (that means this is not a RPN), all subnets
            # should be passed in "subnets" + "extra_subnets".
            if not plugin_loaded or subnets[0]['segment_id'] is None:
                reference['extra_subnets'].append(
                    {k: subnets[2][k] for k in keys})
            self.assertEqual([reference], ports)

    def test__get_sync_floating_ips_no_query(self):
        """Basic test that no query is performed if no router ids are passed"""
        db = l3_db.L3_NAT_dbonly_mixin()
        context = mock.Mock()
        db._get_sync_floating_ips(context, [])
        self.assertFalse(context.session.query.called)

    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin, '_get_floatingip')
    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin, '_is_fip_qos_supported')
    @mock.patch.object(l3_obj.FloatingIP, 'update')
    @mock.patch.object(l3_obj.FloatingIP, 'get_object')
    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin, '_make_floatingip_dict')
    def test__update_floatingip_no_update_existing_qos(self, make_fip_dict,
                                                       mock_get_fip_ovo,
                                                       mock_update_fip_ovo,
                                                       mock_qos_support,
                                                       mock_get_fip):
        fip_id = uuidutils.generate_uuid()
        qos_p_id = uuidutils.generate_uuid()
        db = l3_db.L3_NAT_dbonly_mixin()
        contxt = mock.Mock()

        mock_qos_support.return_value = True

        mock_get_fip.return_value = l3_obj.FloatingIP(
            id=fip_id, qos_policy_id=qos_p_id
        )
        mock_update_fip_ovo.return_value = None
        mock_get_fip_ovo.return_value = l3_obj.FloatingIP(
            id=fip_id, qos_policy_id=qos_p_id
        )
        make_fip_dict.return_value = {
            'id': mock.sentinel.fip_ip,
            'qos_policy_id': qos_p_id
        }

        new_fip = {'floatingip': {'name': 'new_name'}}
        old_fip, fip_dict = db._update_floatingip(contxt, fip_id, new_fip)
        self.assertEqual(qos_p_id, fip_dict['qos_policy_id'])

    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin, '_get_floatingip')
    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin, '_is_fip_qos_supported')
    @mock.patch.object(l3_obj.FloatingIP, 'update')
    @mock.patch.object(l3_obj.FloatingIP, 'get_object')
    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin, '_make_floatingip_dict')
    def test__update_floatingip_update_existing_qos(self, make_fip_dict,
                                                    mock_get_fip_ovo,
                                                    mock_update_fip_ovo,
                                                    mock_qos_support,
                                                    mock_get_fip):
        fip_id = uuidutils.generate_uuid()
        qos_p_id = uuidutils.generate_uuid()
        db = l3_db.L3_NAT_dbonly_mixin()
        contxt = context.get_admin_context()
        mock_qos_support.return_value = True

        fip_ovo = l3_obj.FloatingIP(
            id=fip_id, qos_policy_id=qos_p_id
        )

        mock_get_fip.return_value = fip_ovo
        mock_get_fip_ovo.return_value = fip_ovo
        make_fip_dict.return_value = {
            'id': mock.sentinel.fip_ip,
            'qos_policy_id': qos_p_id
        }

        # Update the QoS Id with a new one
        new_qos_id = uuidutils.generate_uuid()
        new_fip = {'floatingip': {'qos_policy_id': new_qos_id}}
        db._update_floatingip(contxt, fip_id, new_fip)

        mock_update_fip_ovo.assert_called_once()
        self.assertEqual(2, make_fip_dict.call_count)

        # Remove the QoS Id (update to None):
        make_fip_dict.reset_mock()
        mock_update_fip_ovo.reset_mock()
        new_fip = {'floatingip': {'qos_policy_id': None}}
        db._update_floatingip(contxt, fip_id, new_fip)

        mock_update_fip_ovo.assert_called_once()
        self.assertEqual(2, make_fip_dict.call_count)

    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin, '_make_floatingip_dict')
    def test__make_floatingip_dict_with_scope(self, make_fip_dict):
        db = l3_db.L3_NAT_dbonly_mixin()
        make_fip_dict.return_value = {'id': mock.sentinel.fip_ip}
        result = db._make_floatingip_dict_with_scope(
            mock.sentinel.floating_ip_db, mock.sentinel.address_scope_id)
        self.assertEqual({
            'fixed_ip_address_scope': mock.sentinel.address_scope_id,
            'id': mock.sentinel.fip_ip}, result)

    @mock.patch.object(directory, 'get_plugin')
    def test_prevent_l3_port_deletion_port_not_found(self, gp):
        # port not found doesn't prevent
        gp.return_value.get_port.side_effect = n_exc.PortNotFound(port_id='1')
        self.db.prevent_l3_port_deletion(None, None)

    @mock.patch.object(directory, 'get_plugin')
    def test_prevent_l3_port_device_owner_not_router(self, gp):
        # ignores other device owners
        gp.return_value.get_port.return_value = {'device_owner': 'cat'}
        self.db.prevent_l3_port_deletion(None, None)

    @mock.patch.object(directory, 'get_plugin')
    def test_prevent_l3_port_no_fixed_ips(self, gp):
        # without fixed IPs is allowed
        gp.return_value.get_port.return_value = {
            'device_owner': n_const.DEVICE_OWNER_ROUTER_INTF, 'fixed_ips': [],
            'device_id': '44', 'id': 'f',
        }
        with testtools.ExpectedException(n_exc.ServicePortInUse):
            self.db.prevent_l3_port_deletion(mock.Mock(), None)

    @mock.patch.object(directory, 'get_plugin')
    def test_prevent_l3_port_no_router(self, gp):
        # without router is allowed
        gp.return_value.get_port.return_value = {
            'device_owner': n_const.DEVICE_OWNER_ROUTER_INTF,
            'device_id': '44', 'id': 'f',
            'fixed_ips': [{'ip_address': '1.1.1.1', 'subnet_id': '4'}]}
        with mock.patch.object(l3_obj.Router, 'objects_exist',
                               return_value=False):
            self.db.prevent_l3_port_deletion(mock.Mock(), None)

    @mock.patch.object(directory, 'get_plugin')
    def test_prevent_l3_port_existing_router(self, gp):
        gp.return_value.get_port.return_value = {
            'device_owner': n_const.DEVICE_OWNER_ROUTER_INTF,
            'device_id': 'some_router', 'id': 'f',
            'fixed_ips': [{'ip_address': '1.1.1.1', 'subnet_id': '4'}]}
        self.db.get_router = mock.Mock()
        with testtools.ExpectedException(n_exc.ServicePortInUse):
            self.db.prevent_l3_port_deletion(mock.Mock(), None)

    @mock.patch.object(directory, 'get_plugin')
    def test_prevent_l3_port_existing_floating_ip(self, gp):
        ctx = context.get_admin_context()
        gp.return_value.get_port.return_value = {
            'device_owner': n_const.DEVICE_OWNER_FLOATINGIP,
            'device_id': 'some_flip', 'id': 'f',
            'fixed_ips': [{'ip_address': '1.1.1.1', 'subnet_id': '4'}]}
        with mock.patch.object(l3_obj.FloatingIP, 'objects_exist',
                               return_value=mock.Mock()),\
                testtools.ExpectedException(n_exc.ServicePortInUse):

            self.db.prevent_l3_port_deletion(ctx, None)

    @mock.patch.object(l3_obj.FloatingIP, 'objects_exist')
    @mock.patch.object(l3_obj.FloatingIP, 'get_objects')
    def test_prevent_internal_ip_change_for_fip(self,
                                                get_objects,
                                                objects_exist):
        ctx = context.get_admin_context()
        port_id = 'test_internal_port'
        new_fixed_ips = [{'subnet_id': 'test_subnet',
                          'ip_address': '192.168.2.110'}]
        fip_obj_dict = {'fixed_ip_address': '192.168.2.120',
                        'id': 'floating_ip1',
                        'port_id': port_id}
        fip_obj = mock.Mock(**fip_obj_dict)
        objects_exist.return_value = True
        get_objects.return_value = [fip_obj]
        with testtools.ExpectedException(n_exc.BadRequest):
            self.db.prevent_internal_ip_change_for_fip(ctx, port_id,
                                                       new_fixed_ips)

    @mock.patch.object(l3_obj.FloatingIP, 'objects_exist')
    @mock.patch.object(l3_obj.FloatingIP, 'get_objects')
    def test_disassociate_floatingips_conflict_by_fip_attached(self,
                                                               get_objects,
                                                               objects_exist):
        context_tenant = context.Context('tenant', 'tenant', is_admin=False)
        objects_exist.return_value = True
        get_objects.side_effect = [
            [],
            [{'id': 'floating_ip1', 'port_id': 'port_id'}]]
        self.assertRaises(l3_db.FipAssociated,
                          self.db.disassociate_floatingips,
                          context_tenant,
                          'port_id')
        objects_exist.assert_called_once_with(
            mock.ANY, fixed_port_id='port_id')
        expected_calls = [
                mock.call(context_tenant, fixed_port_id='port_id'),
                mock.call(mock.ANY, fixed_port_id='port_id')]
        get_objects.assert_has_calls(expected_calls)

    @mock.patch.object(directory, 'get_plugin')
    def test_subscribe_address_scope_of_subnetpool(self, gp):
        l3_db.L3RpcNotifierMixin()
        registry.publish(resources.SUBNETPOOL_ADDRESS_SCOPE,
                         events.AFTER_UPDATE, mock.ANY,
                         payload=events.DBEventPayload(
                             mock.MagicMock(), resource_id='fake_id'))
        self.assertTrue(gp.return_value.notify_routers_updated.called)

    def test__check_and_get_fip_assoc_with_extra_association_no_change(self):
        fip = {'extra_key': 'value'}
        context = mock.MagicMock()
        floatingip_obj = l3_obj.FloatingIP(
            context,
            id=uuidutils.generate_uuid(),
            floating_network_id=uuidutils.generate_uuid(),
            floating_ip_address=netaddr.IPAddress('8.8.8.8'),
            fixed_port_id=uuidutils.generate_uuid(),
            floating_port_id=uuidutils.generate_uuid())
        with mock.patch.object(
                l3_db.L3_NAT_dbonly_mixin,
                '_get_assoc_data',
                return_value=('1', '2', '3')) as mock_get_assoc_data:
            self.db._check_and_get_fip_assoc(context, fip, floatingip_obj)
            context.session.query.assert_not_called()
            mock_get_assoc_data.assert_called_once_with(
                mock.ANY, fip, floatingip_obj)

    def test__notify_attaching_interface(self):
        with mock.patch.object(l3_db.registry, 'publish') as mock_notify:
            context = mock.MagicMock()
            router_id = 'router_id'
            net_id = 'net_id'
            router_db = mock.Mock()
            router_db.id = router_id
            port = {'network_id': net_id}
            intf = {}
            self.db._notify_attaching_interface(context, router_db, port, intf)

            mock_notify.assert_called_once_with(
                resources.ROUTER_INTERFACE, events.BEFORE_CREATE, self.db,
                payload=mock.ANY)
            payload = mock_notify.mock_calls[0][2]['payload']
            self.assertEqual(context, payload.context)
            self.assertEqual(router_id, payload.resource_id)
            self.assertEqual(net_id, payload.metadata.get('network_id'))
            self.assertEqual(intf, payload.metadata.get('interface_info'))
            self.assertEqual(router_db, payload.latest_state)
            self.assertEqual(port, payload.metadata.get('port'))

    def test__create_gw_port(self):
        # NOTE(slaweq): this test is probably wrong
        # returning dict as gw_port breaks test later in L334 in
        # neutron.db.l3_db file
        router_id = '2afb8434-7380-43a2-913f-ba3a5ad5f349'
        router = l3_models.Router(id=router_id)
        new_network_id = 'net-id'
        ext_ips = [{'subnet_id': 'subnet-id', 'ip_address': '1.1.1.1'}]
        gw_port = {'fixed_ips': [{'subnet_id': 'subnet-id',
                                  'ip_address': '1.1.1.1'}],
                   'id': '8742d007-6f05-4b7e-abdb-11818f608959'}
        ctx = context.get_admin_context()

        with db_api.CONTEXT_WRITER.using(ctx):
            with mock.patch.object(directory, 'get_plugin') as get_p, \
                    mock.patch.object(get_p(), 'get_subnets_by_network',
                                      return_value=mock.ANY), \
                    mock.patch.object(get_p(), '_get_port',
                                      return_value=gw_port), \
                    mock.patch.object(l3_db.L3_NAT_dbonly_mixin,
                                      '_check_for_dup_router_subnets') as \
                    cfdrs, \
                    mock.patch.object(plugin_utils, 'create_port',
                                      return_value=gw_port), \
                    mock.patch.object(ctx.session, 'add'), \
                    mock.patch.object(base_obj.NeutronDbObject, 'create'), \
                    mock.patch.object(l3_db.registry, 'publish') as \
                    mock_notify, \
                    mock.patch.object(l3_db.L3_NAT_dbonly_mixin, '_get_router',
                                      return_value=router):

                self.db._create_gw_port(ctx, router_id=router_id,
                                        router=router,
                                        new_network_id=new_network_id,
                                        ext_ips=ext_ips)

                expected_gw_ips = ['1.1.1.1']

                self.assertTrue(cfdrs.called)
                mock_notify.assert_called_with(
                    resources.ROUTER_GATEWAY, events.AFTER_CREATE,
                    self.db._create_gw_port, payload=mock.ANY)
                cb_payload = mock_notify.mock_calls[1][2]['payload']
                self.assertEqual(ctx, cb_payload.context)
                self.assertEqual(expected_gw_ips,
                                 cb_payload.metadata.get('gateway_ips'))
                self.assertEqual(new_network_id,
                                 cb_payload.metadata.get('network_id'))
                self.assertEqual(router_id, cb_payload.resource_id)

    def _create_router(self, gw_port=True, num_ports=2, create_routes=True):
        # GW CIDR: 10.0.0.0/24
        # Interface CIDRS: 10.0.1.0/24, 10.0.2.0/24, etc.
        router_id = uuidutils.generate_uuid()
        port_gw_cidr = netaddr.IPNetwork('10.0.0.0/24')
        rports = []
        if gw_port:
            port_gw = models_v2.Port(
                id=uuidutils.generate_uuid(),
                fixed_ips=[models_v2.IPAllocation(
                    ip_address=str(port_gw_cidr.ip + 1))])
            rports.append(l3_models.RouterPort(router_id=router_id,
                                               port=port_gw))
        else:
            port_gw = None

        port_cidrs = []
        port_subnets = []
        for idx in range(num_ports):
            cidr = port_gw_cidr.cidr.next(idx + 1)
            port = models_v2.Port(
                id=uuidutils.generate_uuid(),
                fixed_ips=[models_v2.IPAllocation(
                    ip_address=str(cidr.ip + 1))])
            port_cidrs.append(cidr)
            rports.append(l3_models.RouterPort(router_id=router_id, port=port))
            port_subnets.append({'cidr': str(cidr)})

        routes = []
        if create_routes:
            for cidr in [*port_cidrs, port_gw_cidr]:
                routes.append(l3_models.RouterRoute(
                    destination=str(cidr.next(100)),
                    nexthop=str(cidr.ip + 10)))
        return (l3_models.Router(
            id=router_id, attached_ports=rports, route_list=routes,
            gw_port_id=port_gw.id if port_gw else None), port_subnets)

    def test__validate_gw_info(self):
        gw_network = mock.Mock(subnets=[mock.Mock(cidr='10.0.0.0/24')],
                               external=True)
        router, port_subnets = self._create_router(gw_port=False)
        info = {'network_id': 'net_id'}
        with mock.patch.object(self.db._core_plugin, '_get_network',
                               return_value=gw_network), \
                mock.patch.object(self.db._core_plugin, 'get_subnet',
                                  side_effect=port_subnets):
            self.assertEqual(
                'net_id',
                self.db._validate_gw_info(self.ctx, info, [], router))

    def test__validate_gw_info_no_route_connectivity(self):
        gw_network = mock.Mock(subnets=[mock.Mock(cidr='10.50.0.0/24')],
                               external=True)
        router, port_subnets = self._create_router(gw_port=False)
        info = {'network_id': 'net_id'}
        with mock.patch.object(self.db._core_plugin, '_get_network',
                               return_value=gw_network), \
                mock.patch.object(self.db._core_plugin, 'get_subnet',
                                  side_effect=port_subnets):
            self.assertRaises(
                xroute_exc.InvalidRoutes, self.db._validate_gw_info, self.ctx,
                info, [], router)

    def test__validate_gw_info_delete_gateway(self):
        router, port_subnets = self._create_router()
        info = {'network_id': None}
        with mock.patch.object(self.db._core_plugin, '_get_network',
                               return_value=None), \
                mock.patch.object(self.db._core_plugin, 'get_subnet',
                                  side_effect=port_subnets):
            self.assertRaises(
                xroute_exc.InvalidRoutes, self.db._validate_gw_info, self.ctx,
                info, [], router)

    def test__validate_gw_info_delete_gateway_no_route(self):
        gw_network = mock.Mock(subnets=[mock.Mock(cidr='10.50.0.0/24')],
                               external=True)
        router, port_subnets = self._create_router(create_routes=False)
        info = {'network_id': None}
        with mock.patch.object(self.db._core_plugin, '_get_network',
                               return_value=gw_network), \
                mock.patch.object(self.db._core_plugin, 'get_subnet',
                                  side_effect=port_subnets):
            self.assertIsNone(
                self.db._validate_gw_info(mock.ANY, info, [], router))

    def test__raise_on_subnets_overlap_does_not_raise(self):
        subnets = [
            {'id': uuidutils.generate_uuid(),
             'cidr': '10.1.0.0/24'},
            {'id': uuidutils.generate_uuid(),
             'cidr': '10.2.0.0/24'}]
        self.db._raise_on_subnets_overlap(subnets[0], subnets[1])

    def test__raise_on_subnets_overlap_raises(self):
        subnets = [
            {'id': uuidutils.generate_uuid(),
             'cidr': '10.1.0.0/20'},
            {'id': uuidutils.generate_uuid(),
             'cidr': '10.1.10.0/24'}]
        self.assertRaises(
            n_exc.BadRequest, self.db._raise_on_subnets_overlap, subnets[0],
            subnets[1])

    def test__validate_one_router_ipv6_port_per_network(self):
        port = models_v2.Port(
                id=uuidutils.generate_uuid(),
                network_id='foo_network',
                fixed_ips=[models_v2.IPAllocation(
                    ip_address=str(netaddr.IPNetwork(
                        '2001:db8::/32').ip + 1),
                    subnet_id='foo_subnet')])
        rports = [l3_models.RouterPort(router_id='foo_router', port=port)]
        router = l3_models.Router(
            id='foo_router', attached_ports=rports, route_list=[],
            gw_port_id=None)
        new_port = models_v2.Port(
                id=uuidutils.generate_uuid(),
                network_id='foo_network2',
                fixed_ips=[models_v2.IPAllocation(
                    ip_address=str(netaddr.IPNetwork(
                        '2001:db8::/32').ip + 2),
                    subnet_id='foo_subnet')])
        self.db._validate_one_router_ipv6_port_per_network(
            router, new_port)

    def test__validate_one_router_ipv6_port_per_network_mix_ipv4_ipv6(self):
        port = models_v2.Port(
                id=uuidutils.generate_uuid(),
                network_id='foo_network',
                fixed_ips=[models_v2.IPAllocation(
                    ip_address=str(netaddr.IPNetwork(
                        '10.1.10.0/24').ip + 1),
                    subnet_id='foo_subnet')])
        rports = [l3_models.RouterPort(router_id='foo_router', port=port)]
        router = l3_models.Router(
            id='foo_router', attached_ports=rports, route_list=[],
            gw_port_id=None)
        new_port = models_v2.Port(
                id=uuidutils.generate_uuid(),
                network_id='foo_network',
                fixed_ips=[models_v2.IPAllocation(
                    ip_address=str(netaddr.IPNetwork(
                        '2001:db8::/32').ip + 2),
                    subnet_id='foo_subnet')])
        self.db._validate_one_router_ipv6_port_per_network(
            router, new_port)

    def test__validate_one_router_ipv6_port_per_network_distributed_port(self):
        port = models_v2.Port(
                id=uuidutils.generate_uuid(),
                network_id='foo_network',
                device_owner=n_const.DEVICE_OWNER_DVR_INTERFACE,
                fixed_ips=[models_v2.IPAllocation(
                    ip_address=str(netaddr.IPNetwork(
                        '2001:db8::/32').ip + 1),
                    subnet_id='foo_subnet')])
        rports = [l3_models.RouterPort(router_id='foo_router', port=port)]
        router = l3_models.Router(
            id='foo_router', attached_ports=rports, route_list=[],
            gw_port_id=None)
        new_port = models_v2.Port(
                id=uuidutils.generate_uuid(),
                network_id='foo_network',
                device_owner=n_const.DEVICE_OWNER_ROUTER_SNAT,
                fixed_ips=[models_v2.IPAllocation(
                    ip_address=str(netaddr.IPNetwork(
                        '2001:db8::/32').ip + 2),
                    subnet_id='foo_subnet')])
        self.db._validate_one_router_ipv6_port_per_network(router, new_port)

    def test__validate_one_router_ipv6_port_per_network_centralized_snat_port(
            self):
        port = models_v2.Port(
                id=uuidutils.generate_uuid(),
                network_id='foo_network',
                device_owner=n_const.DEVICE_OWNER_ROUTER_SNAT,
                fixed_ips=[models_v2.IPAllocation(
                    ip_address=str(netaddr.IPNetwork(
                        '2001:db8::/32').ip + 1),
                    subnet_id='foo_subnet')])
        rports = [l3_models.RouterPort(router_id='foo_router', port=port)]
        router = l3_models.Router(
            id='foo_router', attached_ports=rports, route_list=[],
            gw_port_id=None)
        new_port = models_v2.Port(
                id=uuidutils.generate_uuid(),
                network_id='foo_network',
                device_owner=n_const.DEVICE_OWNER_DVR_INTERFACE,
                fixed_ips=[models_v2.IPAllocation(
                    ip_address=str(netaddr.IPNetwork(
                        '2001:db8::/32').ip + 2),
                    subnet_id='foo_subnet')])
        self.db._validate_one_router_ipv6_port_per_network(router, new_port)

    def test__validate_one_router_ipv6_port_per_network_failed(self):
        port = models_v2.Port(
                id=uuidutils.generate_uuid(),
                network_id='foo_network',
                fixed_ips=[models_v2.IPAllocation(
                    ip_address=str(netaddr.IPNetwork(
                        '2001:db8::/32').ip + 1),
                    subnet_id='foo_subnet')])
        rports = [l3_models.RouterPort(router_id='foo_router', port=port)]
        router = l3_models.Router(
            id='foo_router', attached_ports=rports, route_list=[],
            gw_port_id=None)
        new_port = models_v2.Port(
                id=uuidutils.generate_uuid(),
                network_id='foo_network',
                fixed_ips=[models_v2.IPAllocation(
                    ip_address=str(netaddr.IPNetwork(
                        '2001:db8::/32').ip + 2),
                    subnet_id='foo_subnet')])
        self.assertRaises(
            n_exc.BadRequest,
            self.db._validate_one_router_ipv6_port_per_network,
            router,
            new_port)


class L3_NAT_db_mixin(base.BaseTestCase):
    def setUp(self):
        super(L3_NAT_db_mixin, self).setUp()
        self.db = l3_db.L3_NAT_db_mixin()

    def _test_create_router(self, external_gateway_info=None):
        router_db = l3_models.Router(id='123')
        router_dict = {'id': '123', 'tenant_id': '456',
                       'external_gateway_info': external_gateway_info}
        # Need to use a copy here as the create_router method pops the gateway
        # information
        router_input = {'router': router_dict.copy()}

        with mock.patch.object(l3_db.L3_NAT_dbonly_mixin, '_create_router_db',
                               return_value=router_db) as crd,\
                mock.patch.object(l3_db.L3_NAT_dbonly_mixin,
                                  '_make_router_dict',
                                  return_value=router_dict),\
                mock.patch.object(l3_db.L3_NAT_dbonly_mixin,
                                  '_update_router_gw_info') as urgi,\
                mock.patch.object(l3_db.L3_NAT_dbonly_mixin, '_get_router',
                                  return_value=router_db),\
                mock.patch.object(l3_db.L3_NAT_db_mixin,
                                  'notify_router_updated') as nru:

            self.db.create_router(mock.Mock(), router_input)
            self.assertTrue(crd.called)
            if external_gateway_info:
                self.assertTrue(urgi.called)
                self.assertTrue(nru.called)
            else:
                self.assertFalse(urgi.called)
                self.assertFalse(nru.called)

    def test_create_router_no_gateway(self):
        self._test_create_router()

    def test_create_router_gateway(self):
        ext_gateway_info = {'network_id': 'net-id', 'enable_snat': True,
                            'external_fixed_ips': [
                                {'subnet_id': 'subnet-id',
                                 'ip_address': 'ip'}]}
        self._test_create_router(ext_gateway_info)

    def test_add_router_interface_no_interface_info(self):
        router_db = l3_models.Router(id='123')
        with mock.patch.object(l3_db.L3_NAT_dbonly_mixin, '_get_router',
                               return_value=router_db):
            self.assertRaises(
                n_exc.BadRequest,
                self.db.add_router_interface, mock.Mock(), router_db.id)


class FakeL3Plugin(l3_db.L3_NAT_dbonly_mixin):
    pass


class L3TestCase(test_db_base_plugin_v2.NeutronDbPluginV2TestCase):

    GET_PORTS_BY_ROUTER_MSG = (
        'The following ports, assigned to router %(router_id)s, do not have a '
        '"routerport" register: %(port_ids)s')

    def setUp(self, *args, **kwargs):
        super(L3TestCase, self).setUp(plugin='ml2')
        self.core_plugin = directory.get_plugin()
        self.ctx = context.get_admin_context()
        self.mixin = FakeL3Plugin()
        directory.add_plugin(plugin_constants.L3, self.mixin)
        self.network = self.create_network()
        self.subnets = []
        self.subnets.append(self.create_subnet(self.network, '1.1.1.1',
                                               '1.1.1.0/24'))
        self.subnets.append(self.create_subnet(self.network, '1.1.2.1',
                                               '1.1.2.0/24'))
        router = {'router': {'name': 'foo_router', 'admin_state_up': True,
                             'tenant_id': 'foo_tenant'}}
        self.router = self.create_router(router)
        self.ports = []
        for subnet in self.subnets:
            ipa = str(netaddr.IPNetwork(subnet['subnet']['cidr']).ip + 10)
            fixed_ips = [{'subnet_id': subnet['subnet']['id'],
                          'ip_address': ipa}]
            self.ports.append(self.create_port(
                self.network['network']['id'], {'fixed_ips': fixed_ips}))
        self.addCleanup(self._clean_objs)

    def _clean_objs(self):
        port_obj.Port.delete_objects(
            self.ctx, network_id=self.network['network']['id'])
        subnet_obj.Subnet.delete_objects(
            self.ctx, network_id=self.network['network']['id'])
        network_obj.Network.get_object(
            self.ctx, id=self.network['network']['id']).delete()
        router_ports = l3_obj.RouterPort.get_objects(
            self.ctx, **{'router_id': self.router['id']})
        for router_port in router_ports:
            router_port.delete()
        l3_obj.Router.get_object(self.ctx, id=self.router['id']).delete()

    def create_router(self, router):
        with db_api.CONTEXT_WRITER.using(self.ctx):
            return self.mixin.create_router(self.ctx, router)

    def create_port(self, net_id, port_info):
        with db_api.CONTEXT_WRITER.using(self.ctx):
            return self._make_port(self.fmt, net_id, **port_info)

    def create_network(self, name=None, **kwargs):
        name = name or 'network1'
        with db_api.CONTEXT_WRITER.using(self.ctx):
            return self._make_network(self.fmt, name, True, **kwargs)

    def create_subnet(self, network, gateway, cidr, **kwargs):
        with db_api.CONTEXT_WRITER.using(self.ctx):
            return self._make_subnet(self.fmt, network, gateway, cidr,
                                     **kwargs)

    def _add_router_interfaces(self):
        return [self.mixin.add_router_interface(
            self.ctx, self.router['id'],
            interface_info={'port_id': port['port']['id']})
            for port in self.ports]

    def _check_routerports(self, ri_statuses):
        port_ids = []
        for idx, ri_status in enumerate(ri_statuses):
            rp_obj = l3_obj.RouterPort.get_object(
                self.ctx, port_id=self.ports[idx]['port']['id'],
                router_id=self.router['id'])
            if ri_status:
                self.assertEqual(self.ports[idx]['port']['id'], rp_obj.port_id)
                port_ids.append(rp_obj.port_id)
            else:
                self.assertIsNone(rp_obj)

        _router_obj = l3_obj.Router.get_object(self.ctx, id=self.router['id'])
        router_port_ids = [rp.port_id for rp in
                           _router_obj.db_obj.attached_ports]
        self.assertEqual(sorted(port_ids), sorted(router_port_ids))

    @mock.patch.object(port_obj, 'LOG')
    def test_remove_router_interface_by_port(self, mock_log):
        self._add_router_interfaces()
        self._check_routerports((True, True))

        interface_info = {'port_id': self.ports[0]['port']['id']}
        self.mixin.remove_router_interface(self.ctx, self.router['id'],
                                           interface_info)
        mock_log.warning.assert_not_called()
        self._check_routerports((False, True))

    @mock.patch.object(port_obj, 'LOG')
    def test_remove_router_interface_by_port_removed_rport(self, mock_log):
        self._add_router_interfaces()
        self._check_routerports((True, True))

        rp_obj = l3_obj.RouterPort.get_object(
            self.ctx, router_id=self.router['id'],
            port_id=self.ports[0]['port']['id'])
        rp_obj.delete()

        interface_info = {'port_id': self.ports[0]['port']['id']}
        self.mixin.remove_router_interface(self.ctx, self.router['id'],
                                           interface_info)
        msg_vars = {'router_id': self.router['id'],
                    'port_ids': {self.ports[0]['port']['id']}}
        mock_log.warning.assert_called_once_with(self.GET_PORTS_BY_ROUTER_MSG,
                                                 msg_vars)
        self._check_routerports((False, True))

    @mock.patch.object(port_obj, 'LOG')
    def test_remove_router_interface_by_subnet(self, mock_log):
        self._add_router_interfaces()
        self._check_routerports((True, True))

        interface_info = {'subnet_id': self.subnets[1]['subnet']['id']}
        self.mixin.remove_router_interface(self.ctx, self.router['id'],
                                           interface_info)
        mock_log.warning.not_called_once()
        self._check_routerports((True, False))

    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin,
                       '_check_for_dup_router_subnets')
    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin,
                       '_raise_on_subnets_overlap')
    def test_add_router_interface_by_port_overlap_detected(
            self, mock_raise_on_subnets_overlap, mock_check_dup):
        # NOTE(froyo): On a normal behaviour this overlapping would be detected
        # by _check_for_dup_router_subnets, in order to evalue the code
        # implemented to cover the race condition when two ports are added
        # simultaneously using colliding cidrs we need to "fake" this method
        # to overpass it and check we achieve the code part that cover the case
        mock_check_dup.return_value = True
        network2 = self.create_network('network2')
        subnet = self.create_subnet(network2, '1.1.1.1', '1.1.1.0/24')
        ipa = str(netaddr.IPNetwork(subnet['subnet']['cidr']).ip + 10)
        fixed_ips = [{'subnet_id': subnet['subnet']['id'], 'ip_address': ipa}]
        port = self.create_port(
                network2['network']['id'], {'fixed_ips': fixed_ips})
        self.mixin.add_router_interface(
            self.ctx, self.router['id'],
            interface_info={'port_id': port['port']['id']})
        mock_raise_on_subnets_overlap.assert_not_called()
        self.mixin.add_router_interface(
            self.ctx, self.router['id'],
            interface_info={'port_id': self.ports[0]['port']['id']})
        mock_raise_on_subnets_overlap.assert_called_once()

    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin,
                       '_check_for_dup_router_subnets')
    @mock.patch.object(l3_db.L3_NAT_dbonly_mixin,
                       '_raise_on_subnets_overlap')
    def test_add_router_interface_by_subnet_overlap_detected(
            self, mock_raise_on_subnets_overlap, mock_check_dup):
        # NOTE(froyo): On a normal behaviour this overlapping would be detected
        # by _check_for_dup_router_subnets, in order to evalue the code
        # implemented to cover the race condition when two ports are added
        # simultaneously using colliding cidrs we need to "fake" this method
        # to overpass it and check we achieve the code part that cover the case
        mock_check_dup.return_value = True
        network2 = self.create_network('network2')
        subnet = self.create_subnet(network2, '1.1.1.1', '1.1.1.0/24')
        self.mixin.add_router_interface(
            self.ctx, self.router['id'],
            interface_info={'subnet_id': subnet['subnet']['id']})
        mock_raise_on_subnets_overlap.assert_not_called()
        self.mixin.add_router_interface(
            self.ctx, self.router['id'],
            interface_info={'subnet_id': self.subnets[0]['subnet']['id']})
        mock_raise_on_subnets_overlap.assert_called_once()

    @mock.patch.object(port_obj, 'LOG')
    def test_remove_router_interface_by_subnet_removed_rport(self, mock_log):
        self._add_router_interfaces()
        self._check_routerports((True, True))

        rp_obj = l3_obj.RouterPort.get_object(
            self.ctx, router_id=self.router['id'],
            port_id=self.ports[0]['port']['id'])
        rp_obj.delete()

        interface_info = {'subnet_id': self.subnets[0]['subnet']['id']}
        self.mixin.remove_router_interface(self.ctx, self.router['id'],
                                           interface_info)
        msg_vars = {'router_id': self.router['id'],
                    'port_ids': {self.ports[0]['port']['id']}}
        mock_log.warning.assert_called_once_with(self.GET_PORTS_BY_ROUTER_MSG,
                                                 msg_vars)
        self._check_routerports((False, True))

    def test_create_router_notify(self):
        with mock.patch.object(l3_db.registry, 'publish') as mock_publish:
            router = {'router': {'name': 'foo_router',
                                 'admin_state_up': True,
                                 'tenant_id': 'foo_tenant'}}
            self.create_router(router)
            expected_calls = [
                mock.call(resources.ROUTER, events.BEFORE_CREATE,
                          self.mixin, payload=mock.ANY),
                mock.call(resources.ROUTER, events.PRECOMMIT_CREATE,
                          self.mixin, payload=mock.ANY),
                mock.call(resources.ROUTER, events.AFTER_CREATE,
                          self.mixin, payload=mock.ANY),
            ]
            mock_publish.assert_has_calls(expected_calls)

    def test_create_router_extra_attr(self):
        router_args = {'router': {'name': 'foo_router',
                                  'admin_state_up': True,
                                  'tenant_id': 'foo_tenant'}
                       }
        router_dict = self.create_router(router_args)
        with db_api.CONTEXT_READER.using(self.ctx) as session:
            r_extra_attrs = session.query(
                l3_attrs.RouterExtraAttributes).filter(
                    l3_attrs.RouterExtraAttributes.router_id ==
                    router_dict['id']).all()
        self.assertEqual(1, len(r_extra_attrs))
        self.assertEqual(router_dict['id'], r_extra_attrs[0].router_id)

    def test_update_router_notify(self):
        with mock.patch.object(l3_db.registry, 'publish') as mock_publish:
            self.mixin.update_router(self.ctx, self.router['id'],
                                     {'router': {'name': 'test1'}})
            expected_calls = [
                mock.call(resources.ROUTER, events.PRECOMMIT_UPDATE,
                          self.mixin, payload=mock.ANY),
                mock.call(resources.ROUTER, events.AFTER_UPDATE,
                         self.mixin, payload=mock.ANY),
            ]
            mock_publish.assert_has_calls(expected_calls)

    def _create_external_network(self, name=None, **kwargs):
        name = name or 'network1'
        kwargs[extnet_apidef.EXTERNAL] = True
        with db_api.CONTEXT_WRITER.using(self.ctx):
            res = self._create_network(
                self.fmt, name, True,
                arg_list=(extnet_apidef.EXTERNAL,),
                as_admin=True, **kwargs)
            if res.status_int >= webob.exc.HTTPClientError.code:
                raise webob.exc.HTTPClientError(code=res.status_int)
            return self.deserialize(self.fmt, res)

    def test_update_router_gw_notify(self):
        with mock.patch.object(l3_db.registry, 'publish') as mock_publish:
            ext_net = self._create_external_network()
            self.create_subnet(ext_net, '1.1.2.1', '1.1.2.0/24')
            update_data = {
                l3_apidef.EXTERNAL_GW_INFO: {
                    'network_id': ext_net['network']['id']}}
            self.mixin.update_router(
                self.ctx, self.router['id'], {'router': update_data})
            expected_calls = [
                mock.call(resources.NETWORK, events.BEFORE_CREATE,
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.SEGMENT, events.PRECOMMIT_CREATE,
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.NETWORK, events.PRECOMMIT_CREATE,
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.NETWORK, events.AFTER_CREATE,
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.NETWORK, events.BEFORE_RESPONSE,
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.SUBNET, events.BEFORE_CREATE,
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.SUBNET, events.AFTER_CREATE,
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.SUBNET, events.BEFORE_RESPONSE,
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.ROUTER_GATEWAY, events.BEFORE_CREATE,
                          self.mixin, payload=mock.ANY),
                mock.call(resources.PORT, events.BEFORE_CREATE,
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.PORT, events.PRECOMMIT_CREATE,
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.PORT, events.AFTER_CREATE,
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.ROUTER_GATEWAY, events.AFTER_CREATE,
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.ROUTER, events.PRECOMMIT_UPDATE,
                          self.mixin, payload=mock.ANY),
                mock.call(resources.ROUTER, events.AFTER_UPDATE,
                          self.mixin, payload=mock.ANY)]
            mock_publish.assert_has_calls(expected_calls)
            mock_publish.reset_mock()
            update_data = {l3_apidef.EXTERNAL_GW_INFO: {}}
            self.mixin.update_router(
                self.ctx, self.router['id'], {'router': update_data})
            expected_calls = [
                mock.call(resources.ROUTER_GATEWAY, events.BEFORE_DELETE,
                          self.mixin, payload=mock.ANY),
                mock.call(resources.PORT, events.BEFORE_DELETE,
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.PORT, events.PRECOMMIT_DELETE,
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.PORT, events.AFTER_DELETE,
                          mock.ANY, payload=mock.ANY),
                mock.call(resources.ROUTER_GATEWAY, events.AFTER_DELETE,
                          self.mixin, payload=mock.ANY),
                mock.call(resources.ROUTER, events.PRECOMMIT_UPDATE,
                          self.mixin, payload=mock.ANY),
                mock.call(resources.ROUTER, events.AFTER_UPDATE,
                          self.mixin, payload=mock.ANY)]
            mock_publish.assert_has_calls(expected_calls)



# Copyright (c) 2021 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslo_policy import policy as base_policy
from oslo_utils import uuidutils

from neutron import policy
from neutron.tests.unit.conf.policies import test_base as base


class PortAPITestCase(base.PolicyBaseTestCase):

    def setUp(self):
        super(PortAPITestCase, self).setUp()

        self.network = {
            'id': uuidutils.generate_uuid(),
            'project_id': self.project_id}
        self.target = {
            'project_id': self.project_id,
            'tenant_id': self.alt_project_id,
            'network_id': self.network['id'],
            'ext_parent_network_id': self.network['id']}
        self.alt_target = {
            'project_id': self.alt_project_id,
            'tenant_id': self.alt_project_id,
            'network_id': self.network['id'],
            'ext_parent_network_id': self.network['id']}

        self.plugin_mock = mock.Mock()
        self.plugin_mock.get_network.return_value = self.network
        mock.patch(
            'neutron_lib.plugins.directory.get_plugin',
            return_value=self.plugin_mock).start()


class SystemAdminTests(PortAPITestCase):

    def setUp(self):
        super(SystemAdminTests, self).setUp()
        self.context = self.system_admin_ctx

    def test_create_port(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port', self.alt_target)

    def test_create_port_with_device_owner(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:device_owner',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:device_owner',
            self.alt_target)

    def test_create_port_with_mac_address(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:mac_address',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:mac_address',
            self.alt_target)

    def test_create_port_with_fixed_ips(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:fixed_ips',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:fixed_ips',
            self.alt_target)

    def test_create_port_with_fixed_ips_and_ip_address(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:fixed_ips:ip_address',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:fixed_ips:ip_address',
            self.alt_target)

    def test_create_port_with_fixed_ips_and_subnet_id(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:fixed_ips:subnet_id',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:fixed_ips:subnet_id',
            self.alt_target)

    def test_create_port_with_port_security_enabled(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:port_security_enabled',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:port_security_enabled',
            self.alt_target)

    def test_create_port_with_binding_host_id(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:binding:host_id',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:binding:host_id',
            self.alt_target)

    def test_create_port_with_binding_profile(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:binding:profile',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:binding:profile',
            self.alt_target)

    def test_create_port_with_binding_vnic_type(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:binding:vnic_type',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'create_port:binding:vnic_type',
            self.alt_target)

    def test_create_port_with_allowed_address_pairs(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_port:allowed_address_pairs',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_port:allowed_address_pairs',
            self.alt_target)

    def test_create_port_with_allowed_address_pairs_and_mac_address(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_port:allowed_address_pairs:mac_address',
            self.alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_port:allowed_address_pairs:mac_address',
            self.target)

    def test_create_port_with_allowed_address_pairs_and_ip_address(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_port:allowed_address_pairs:ip_address',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_port:allowed_address_pairs:ip_address',
            self.alt_target)

    def test_get_port(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'get_port', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'get_port', self.alt_target)

    def test_get_port_binding_vif_type(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'get_port:binding:vif_type',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'get_port:binding:vif_type',
            self.alt_target)

    def test_get_port_binding_vif_details(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'get_port:binding:vif_details',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'get_port:binding:vif_details',
            self.alt_target)

    def test_get_port_binding_host_id(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'get_port:binding:host_id',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'get_port:binding:host_id',
            self.alt_target)

    def test_get_port_binding_profile(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'get_port:binding:profile',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'get_port:binding:profile',
            self.alt_target)

    def test_get_port_resource_request(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'get_port:resource_request',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'get_port:resource_request',
            self.alt_target)

    def test_get_ports_tags(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'get_ports_tags', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'get_ports_tags', self.alt_target)

    def test_update_port(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port', self.alt_target)

    def test_update_port_with_device_owner(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:device_owner',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:device_owner',
            self.alt_target)

    def test_update_port_with_mac_address(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:mac_address',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:mac_address',
            self.alt_target)

    def test_update_port_with_fixed_ips(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:fixed_ips',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:fixed_ips',
            self.alt_target)

    def test_update_port_with_fixed_ips_and_ip_address(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:fixed_ips:ip_address',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:fixed_ips:ip_address',
            self.alt_target)

    def test_update_port_with_fixed_ips_and_subnet_id(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:fixed_ips:subnet_id',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:fixed_ips:subnet_id',
            self.alt_target)

    def test_update_port_with_port_security_enabled(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:port_security_enabled',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:port_security_enabled',
            self.alt_target)

    def test_update_port_with_binding_host_id(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:binding:host_id',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:binding:host_id',
            self.alt_target)

    def test_update_port_with_binding_profile(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:binding:profile',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:binding:profile',
            self.alt_target)

    def test_update_port_with_binding_vnic_type(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:binding:vnic_type',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'update_port:binding:vnic_type',
            self.alt_target)

    def test_update_port_with_allowed_address_pairs(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_port:allowed_address_pairs',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_port:allowed_address_pairs',
            self.alt_target)

    def test_update_port_with_allowed_address_pairs_and_mac_address(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_port:allowed_address_pairs:mac_address',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_port:allowed_address_pairs:mac_address',
            self.alt_target)

    def test_update_port_with_allowed_address_pairs_and_ip_address(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_port:allowed_address_pairs:ip_address',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_port:allowed_address_pairs:ip_address',
            self.alt_target)

    def test_update_port_data_plane_status(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_port:data_plane_status', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_port:data_plane_status', self.alt_target)

    def test_delete_port(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'delete_port', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce, self.context, 'delete_port', self.alt_target)


class SystemMemberTests(SystemAdminTests):

    def setUp(self):
        super(SystemMemberTests, self).setUp()
        self.context = self.system_member_ctx


class SystemReaderTests(SystemMemberTests):

    def setUp(self):
        super(SystemReaderTests, self).setUp()
        self.context = self.system_reader_ctx


class AdminTests(PortAPITestCase):

    def setUp(self):
        super(AdminTests, self).setUp()
        self.context = self.project_admin_ctx

    def test_create_port(self):
        self.assertTrue(
            policy.enforce(self.context, 'create_port', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'create_port', self.alt_target))

    def test_create_port_with_device_owner(self):
        target = self.target.copy()
        target['device_owner'] = 'network:test'
        alt_target = self.alt_target.copy()
        alt_target['device_owner'] = 'network:test'
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:device_owner', target))
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:device_owner', alt_target))

    def test_create_port_with_mac_address(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:mac_address', self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:mac_address', self.alt_target))

    def test_create_port_with_fixed_ips(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:fixed_ips', self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:fixed_ips', self.alt_target))

    def test_create_port_with_fixed_ips_and_ip_address(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:fixed_ips:ip_address', self.target))
        self.assertTrue(
            policy.enforce(
                self.context,
                'create_port:fixed_ips:ip_address', self.alt_target))

    def test_create_port_with_fixed_ips_and_subnet_id(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:fixed_ips:subnet_id', self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:fixed_ips:subnet_id', self.alt_target))

    def test_create_port_with_port_security_enabled(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:port_security_enabled', self.target))
        self.assertTrue(
            policy.enforce(
                self.context,
                'create_port:port_security_enabled', self.alt_target))

    def test_create_port_with_binding_host_id(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:binding:host_id', self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:binding:host_id', self.alt_target))

    def test_create_port_with_binding_profile(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:binding:profile', self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:binding:profile', self.alt_target))

    def test_create_port_with_binding_vnic_type(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:binding:vnic_type', self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:binding:vnic_type', self.alt_target))

    def test_create_port_with_allowed_address_pairs(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:allowed_address_pairs', self.target))
        self.assertTrue(
            policy.enforce(
                self.context,
                'create_port:allowed_address_pairs', self.alt_target))

    def test_create_port_with_allowed_address_pairs_and_mac_address(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:allowed_address_pairs:mac_address',
                           self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:allowed_address_pairs:mac_address',
                           self.alt_target))

    def test_create_port_with_allowed_address_pairs_and_ip_address(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:allowed_address_pairs:ip_address',
                           self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:allowed_address_pairs:ip_address',
                           self.alt_target))

    def test_create_port_with_hints(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:hints',
                           self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:hints',
                           self.alt_target))

    def test_get_port(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_port', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'get_port', self.alt_target))

    def test_get_port_binding_vif_type(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'get_port:binding:vif_type', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'get_port:binding:vif_type', self.alt_target))

    def test_get_port_binding_vif_details(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'get_port:binding:vif_details', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'get_port:binding:vif_details', self.alt_target))

    def test_get_port_binding_host_id(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'get_port:binding:host_id', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'get_port:binding:host_id', self.alt_target))

    def test_get_port_binding_profile(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'get_port:binding:profile', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'get_port:binding:profile', self.alt_target))

    def test_get_port_resource_request(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'get_port:resource_request', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'get_port:resource_request', self.alt_target))

    def test_get_port_hints(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'get_port:hints', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'get_port:hints', self.alt_target))

    def test_get_ports_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_ports_tags', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'get_ports_tags', self.alt_target))

    def test_update_port(self):
        self.assertTrue(
            policy.enforce(self.context, 'update_port', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'update_port', self.alt_target))

    def test_update_port_with_device_owner(self):
        target = self.target.copy()
        target['device_owner'] = 'network:test'
        alt_target = self.alt_target.copy()
        alt_target['device_owner'] = 'network:test'
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:device_owner', target))
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:device_owner', alt_target))

    def test_update_port_with_mac_address(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'update_port:mac_address', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'update_port:mac_address', self.alt_target))

    def test_update_port_with_fixed_ips(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:fixed_ips', self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:fixed_ips', self.alt_target))

    def test_update_port_with_fixed_ips_and_ip_address(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:fixed_ips:ip_address', self.target))
        self.assertTrue(
            policy.enforce(
                self.context,
                'update_port:fixed_ips:ip_address', self.alt_target))

    def test_update_port_with_fixed_ips_and_subnet_id(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:fixed_ips:subnet_id', self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:fixed_ips:subnet_id', self.alt_target))

    def test_update_port_with_port_security_enabled(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:port_security_enabled', self.target))
        self.assertTrue(
            policy.enforce(
                self.context,
                'update_port:port_security_enabled', self.alt_target))

    def test_update_port_with_binding_host_id(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:binding:host_id', self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:binding:host_id', self.alt_target))

    def test_update_port_with_binding_profile(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:binding:profile', self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:binding:profile', self.alt_target))

    def test_update_port_with_binding_vnic_type(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:binding:vnic_type', self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:binding:vnic_type', self.alt_target))

    def test_update_port_with_allowed_address_pairs(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:allowed_address_pairs', self.target))
        self.assertTrue(
            policy.enforce(
                self.context,
                'update_port:allowed_address_pairs', self.alt_target))

    def test_update_port_with_allowed_address_pairs_and_mac_address(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:allowed_address_pairs:mac_address',
                           self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:allowed_address_pairs:mac_address',
                           self.alt_target))

    def test_update_port_with_allowed_address_pairs_and_ip_address(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:allowed_address_pairs:ip_address',
                           self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:allowed_address_pairs:ip_address',
                           self.alt_target))

    def test_update_port_data_plane_status(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:data_plane_status',
                           self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:data_plane_status',
                           self.alt_target))

    def test_update_port_hints(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:hints',
                           self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:hints',
                           self.alt_target))

    def test_delete_port(self):
        self.assertTrue(
            policy.enforce(self.context, 'delete_port', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'delete_port', self.alt_target))


class ProjectMemberTests(AdminTests):

    def setUp(self):
        super(ProjectMemberTests, self).setUp()
        self.context = self.project_member_ctx

    def test_create_port(self):
        self.assertTrue(
            policy.enforce(self.context, 'create_port', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port', self.alt_target)

    def test_create_port_with_device_owner(self):
        target = self.target.copy()
        target['device_owner'] = 'network:test'
        alt_target = self.alt_target.copy()
        alt_target['device_owner'] = 'network:test'
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:device_owner',
            target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:device_owner',
            alt_target)

    def test_create_port_with_mac_address(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:mac_address',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:mac_address',
            self.alt_target)

    def test_create_port_with_fixed_ips(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:fixed_ips',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:fixed_ips',
            self.alt_target)

    def test_create_port_with_fixed_ips_and_ip_address(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:fixed_ips:ip_address',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:fixed_ips:ip_address',
            self.alt_target)

    def test_create_port_with_fixed_ips_and_subnet_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:fixed_ips:subnet_id',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:fixed_ips:subnet_id',
            self.alt_target)

    def test_create_port_with_port_security_enabled(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:port_security_enabled',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:port_security_enabled',
            self.alt_target)

    def test_create_port_with_binding_host_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:binding:host_id',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:binding:host_id',
            self.alt_target)

    def test_create_port_with_binding_profile(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:binding:profile',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:binding:profile',
            self.alt_target)

    def test_create_port_with_binding_vnic_type(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_port:binding:vnic_type', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:binding:vnic_type',
            self.alt_target)

    def test_create_port_with_allowed_address_pairs(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_port:allowed_address_pairs',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_port:allowed_address_pairs',
            self.alt_target)

    def test_create_port_with_allowed_address_pairs_and_mac_address(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_port:allowed_address_pairs:mac_address',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_port:allowed_address_pairs:mac_address',
            self.alt_target)

    def test_create_port_with_allowed_address_pairs_and_ip_address(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_port:allowed_address_pairs:ip_address',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_port:allowed_address_pairs:ip_address',
            self.alt_target)

    def test_create_port_with_hints(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_port:hints',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_port:hints',
            self.alt_target)

    def test_get_port(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_port', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port', self.alt_target)

    def test_get_port_binding_vif_type(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:binding:vif_type',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:binding:vif_type',
            self.alt_target)

    def test_get_port_binding_vif_details(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:binding:vif_details',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:binding:vif_details',
            self.alt_target)

    def test_get_port_binding_host_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:binding:host_id',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:binding:host_id',
            self.alt_target)

    def test_get_port_binding_profile(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:binding:profile',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:binding:profile',
            self.alt_target)

    def test_get_port_resource_request(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:resource_request',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:resource_request',
            self.alt_target)

    def test_get_port_hints(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:hints',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:hints',
            self.alt_target)

    def test_get_ports_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_ports_tags', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_ports_tags', self.alt_target)

    def test_update_port(self):
        self.assertTrue(
            policy.enforce(self.context, 'update_port', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port', self.alt_target)

    def test_update_port_with_device_owner(self):
        target = self.target.copy()
        target['device_owner'] = 'network:test'
        alt_target = self.alt_target.copy()
        alt_target['device_owner'] = 'network:test'
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:device_owner',
            target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:device_owner',
            alt_target)

    def test_update_port_with_mac_address(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:mac_address',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:mac_address',
            self.alt_target)

    def test_update_port_with_fixed_ips(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:fixed_ips',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:fixed_ips',
            self.alt_target)

    def test_update_port_with_fixed_ips_and_ip_address(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:fixed_ips:ip_address',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:fixed_ips:ip_address',
            self.alt_target)

    def test_update_port_with_fixed_ips_and_subnet_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:fixed_ips:subnet_id',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:fixed_ips:subnet_id',
            self.alt_target)

    def test_update_port_with_port_security_enabled(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:port_security_enabled',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:port_security_enabled',
            self.alt_target)

    def test_update_port_with_binding_host_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:binding:host_id',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:binding:host_id',
            self.alt_target)

    def test_update_port_with_binding_profile(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:binding:profile',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:binding:profile',
            self.alt_target)

    def test_update_port_with_binding_vnic_type(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_port:binding:vnic_type', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:binding:vnic_type',
            self.alt_target)

    def test_update_port_with_allowed_address_pairs(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_port:allowed_address_pairs',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_port:allowed_address_pairs',
            self.alt_target)

    def test_update_port_with_allowed_address_pairs_and_mac_address(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_port:allowed_address_pairs:mac_address',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_port:allowed_address_pairs:mac_address',
            self.alt_target)

    def test_update_port_with_allowed_address_pairs_and_ip_address(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_port:allowed_address_pairs:ip_address',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_port:allowed_address_pairs:ip_address',
            self.alt_target)

    def test_update_port_data_plane_status(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_port:data_plane_status', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_port:data_plane_status', self.alt_target)

    def test_update_port_hints(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_port:hints', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_port:hints', self.alt_target)

    def test_update_ports_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'update_ports_tags', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_ports_tags', self.alt_target)

    def test_delete_port(self):
        self.assertTrue(
            policy.enforce(self.context, 'delete_port', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'delete_port', self.alt_target)


class ProjectReaderTests(ProjectMemberTests):

    def setUp(self):
        super(ProjectReaderTests, self).setUp()
        self.context = self.project_reader_ctx

    def test_create_port(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port', self.alt_target)

    def test_create_port_with_binding_vnic_type(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:binding:vnic_type',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:binding:vnic_type',
            self.alt_target)

    def test_update_port(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port', self.alt_target)

    def test_update_port_with_binding_vnic_type(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:binding:vnic_type',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:binding:vnic_type',
            self.alt_target)

    def test_update_ports_tags(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_ports_tags', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_ports_tags', self.alt_target)

    def test_delete_port(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'delete_port', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'delete_port', self.alt_target)


class ServiceRoleTests(PortAPITestCase):

    def setUp(self):
        super(ServiceRoleTests, self).setUp()
        self.context = self.service_ctx

    def test_create_port(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port', self.target)

    def test_create_port_with_device_owner(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'create_port:device_owner', self.target))

    def test_create_port_with_mac_address(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'create_port:mac_address', self.target))

    def test_create_port_with_fixed_ips(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'create_port:fixed_ips', self.target))

    def test_create_port_with_fixed_ips_and_ip_address(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'create_port:fixed_ips:ip_address', self.target))

    def test_create_port_with_fixed_ips_and_subnet_id(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'create_port:fixed_ips:subnet_id', self.target))

    def test_create_port_with_port_security_enabled(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'create_port:port_security_enabled',
                self.target))

    def test_create_port_with_binding_host_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:binding:host_id',
            self.target)

    def test_create_port_with_binding_profile(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:binding:profile',
            self.target)

    def test_create_port_with_binding_vnic_type(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'create_port:binding:vnic_type',
            self.target)

    def test_create_port_with_allowed_address_pairs(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_port:allowed_address_pairs',
            self.target)

    def test_create_port_with_allowed_address_pairs_and_mac_address(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_port:allowed_address_pairs:mac_address',
            self.alt_target)

    def test_create_port_with_allowed_address_pairs_and_ip_address(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_port:allowed_address_pairs:ip_address',
            self.target)

    def test_get_port(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_port', self.target))

    def test_get_port_binding_vif_type(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:binding:vif_type',
            self.target)

    def test_get_port_binding_vif_details(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:binding:vif_details',
            self.target)

    def test_get_port_binding_host_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:binding:host_id',
            self.target)

    def test_get_port_binding_profile(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:binding:profile',
            self.target)

    def test_get_port_resource_request(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'get_port:resource_request',
            self.target)

    def test_update_port(self):
        self.assertTrue(
            policy.enforce(self.context, 'update_port', self.target))

    def test_update_port_with_device_owner(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'update_port:device_owner', self.target))

    def test_update_port_with_mac_address(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'update_port:mac_address', self.target))

    def test_update_port_with_fixed_ips(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'update_port:fixed_ips', self.target))

    def test_update_port_with_fixed_ips_and_ip_address(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'update_port:fixed_ips:ip_address', self.target))

    def test_update_port_with_fixed_ips_and_subnet_id(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'update_port:fixed_ips:subnet_id', self.target))

    def test_update_port_with_port_security_enabled(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'update_port:port_security_enabled',
                self.target))

    def test_update_port_with_binding_host_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:binding:host_id',
            self.target)

    def test_update_port_with_binding_profile(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce, self.context, 'update_port:binding:profile',
            self.target)

    def test_update_port_with_binding_vnic_type(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'update_port:binding:vnic_type', self.target))

    def test_update_port_with_allowed_address_pairs(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_port:allowed_address_pairs',
            self.target)

    def test_update_port_with_allowed_address_pairs_and_mac_address(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_port:allowed_address_pairs:mac_address',
            self.target)

    def test_update_port_with_allowed_address_pairs_and_ip_address(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_port:allowed_address_pairs:ip_address',
            self.target)

    def test_update_port_data_plane_status(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_port:data_plane_status', self.target)

    def test_delete_port(self):
        self.assertTrue(
            policy.enforce(self.context, 'delete_port', self.target))



# Copyright 2016 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from neutron_lib.agent import topics
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_db import options as db_options
from oslo_log import log as logging

from neutron.common import config as common_config
from neutron.conf.agent import securitygroups_rpc
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron import manager
from neutron import opts as neutron_options
from neutron.plugins.ml2.drivers.ovn import db_migration
from neutron.plugins.ml2.drivers.ovn.mech_driver import mech_driver
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import impl_idl_ovn
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import ovn_db_sync
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import worker
from neutron.plugins.ml2 import plugin as ml2_plugin

LOG = logging.getLogger(__name__)


class Ml2Plugin(ml2_plugin.Ml2Plugin):

    def _setup_dhcp(self):
        pass

    def _start_rpc_notifiers(self):
        # Override the notifier so that when calling the ML2 plugin to create
        # resources, it doesn't crash trying to notify subscribers.
        self.notifier = AgentNotifierApi(topics.AGENT)


class OVNMechanismDriver(mech_driver.OVNMechanismDriver):

    def subscribe(self):
        pass

    def post_fork_initialize(self, resource, event, trigger, **kwargs):
        pass

    @property
    def ovn_client(self):
        return self._ovn_client

    def _remove_node_from_hash_ring(self):
        """Don't remove the node from the Hash Ring.

        If this method was not overridden, cleanup would be performed when
        calling the db sync and running neutron server would remove the
        nodes from the Hash Ring.
        """

    # Since we are not using the ovn mechanism driver while syncing,
    # we override the post and pre commit methods so that original ones are
    # not called.
    def create_port_precommit(self, context):
        pass

    def create_port_postcommit(self, context):
        port = context.current
        self.ovn_client.create_port(context.plugin_context, port)

    def update_port_precommit(self, context):
        pass

    def update_port_postcommit(self, context):
        port = context.current
        original_port = context.original
        self.ovn_client.update_port(context.plugin_context, port,
                                    original_port)

    def delete_port_precommit(self, context):
        pass

    def delete_port_postcommit(self, context):
        port = copy.deepcopy(context.current)
        port['network'] = context.network.current
        self.ovn_client.delete_port(context.plugin_context, port['id'])


class AgentNotifierApi(object):
    """Default Agent Notifier class for ovn-db-sync-util.

    This class implements empty methods so that when creating resources in
    the core plugin, the original ones don't get called and don't interfere
    with the syncing process.
    """
    def __init__(self, topic):
        self.topic = topic
        self.topic_network_delete = topics.get_topic_name(topic,
                                                          topics.NETWORK,
                                                          topics.DELETE)
        self.topic_port_update = topics.get_topic_name(topic,
                                                       topics.PORT,
                                                       topics.UPDATE)
        self.topic_port_delete = topics.get_topic_name(topic,
                                                       topics.PORT,
                                                       topics.DELETE)
        self.topic_network_update = topics.get_topic_name(topic,
                                                          topics.NETWORK,
                                                          topics.UPDATE)

    def network_delete(self, context, network_id):
        pass

    def port_update(self, context, port, network_type, segmentation_id,
                    physical_network):
        pass

    def port_delete(self, context, port_id):
        pass

    def network_update(self, context, network):
        pass

    def security_groups_provider_updated(self, context,
                                         devices_to_update=None):
        pass


def setup_conf():
    conf = cfg.CONF
    common_config.register_common_config_options()
    ovn_conf.register_opts()
    ml2_group, ml2_opts = neutron_options.list_ml2_conf_opts()[0]
    cfg.CONF.register_cli_opts(ml2_opts, ml2_group)
    cfg.CONF.register_cli_opts(securitygroups_rpc.security_group_opts,
                               'SECURITYGROUP')
    ovn_group, ovn_opts = ovn_conf.list_opts()[0]
    cfg.CONF.register_cli_opts(ovn_opts, group=ovn_group)
    db_group, neutron_db_opts = db_options.list_opts()[0]
    cfg.CONF.register_cli_opts(neutron_db_opts, db_group)
    # Override Nova notify configuration LP: #1882020
    cfg.CONF.set_override('notify_nova_on_port_status_changes', False)
    cfg.CONF.set_override('notify_nova_on_port_data_changes', False)
    return conf


def main():
    """Main method for syncing neutron networks and ports with ovn nb db.

    This script provides a utility for syncing the OVN Northbound Database
    with the Neutron database.

    This script is used for the migration from ML2/OVS to ML2/OVN.
    """
    conf = setup_conf()

    # if no config file is passed or no configuration options are passed
    # then load configuration from /etc/neutron/neutron.conf
    try:
        conf(project='neutron')
    except TypeError:
        LOG.error('Error parsing the configuration values. Please verify.')
        return

    logging.setup(conf, 'neutron_ovn_db_sync_util', fix_eventlet=False)
    LOG.info('Neutron OVN DB sync started')
    mode = ovn_conf.get_ovn_neutron_sync_mode()
    # Migrate mode will run as repair mode in the synchronizer
    migrate = False
    if mode == ovn_conf.MIGRATE_MODE:
        mode = ovn_db_sync.SYNC_MODE_REPAIR
        migrate = True
    if mode not in [ovn_db_sync.SYNC_MODE_LOG, ovn_db_sync.SYNC_MODE_REPAIR]:
        LOG.error(
            'Invalid sync mode: ["%s"]. Should be "log" or "repair"', mode)
        return

    # Validate and modify core plugin and ML2 mechanism drivers for syncing.
    if (cfg.CONF.core_plugin.endswith('.Ml2Plugin') or
            cfg.CONF.core_plugin == 'ml2'):
        cfg.CONF.core_plugin = (
            'neutron.cmd.ovn.neutron_ovn_db_sync_util.Ml2Plugin')
        if not cfg.CONF.ml2.mechanism_drivers:
            LOG.error('Please use --config-file to specify '
                      'neutron and ml2 configuration file.')
            return
        if 'ovn' not in cfg.CONF.ml2.mechanism_drivers:
            LOG.error('No "ovn" mechanism driver found: "%s".',
                      cfg.CONF.ml2.mechanism_drivers)
            return
        cfg.CONF.set_override('mechanism_drivers', ['ovn-sync'], 'ml2')
        conf.service_plugins = [
            'neutron.services.ovn_l3.plugin.OVNL3RouterPlugin',
            'neutron.services.segments.plugin.Plugin',
            'port_forwarding',
            'qos'
        ]
        extension_drivers = list(set(cfg.CONF.ml2.extension_drivers + ['qos']))
        cfg.CONF.set_override('extension_drivers', extension_drivers, 'ml2')

    else:
        LOG.error('Invalid core plugin: ["%s"].', cfg.CONF.core_plugin)
        return

    mech_worker = worker.MaintenanceWorker
    try:
        ovn_api = impl_idl_ovn.OvsdbNbOvnIdl.from_worker(mech_worker)
    except RuntimeError:
        LOG.error('Invalid --ovn-ovn_nb_connection parameter provided.')
        return

    try:
        ovn_sb_api = impl_idl_ovn.OvsdbSbOvnIdl.from_worker(mech_worker)
    except RuntimeError:
        LOG.error('Invalid --ovn-ovn_sb_connection parameter provided.')
        return

    manager.init()
    core_plugin = directory.get_plugin()
    driver = core_plugin.mechanism_manager.mech_drivers['ovn-sync']
    # The L3 code looks for the OVSDB connection on the 'ovn' driver
    # and will fail with a KeyError if it isn't there
    core_plugin.mechanism_manager.mech_drivers['ovn'] = driver
    ovn_driver = driver.obj
    ovn_driver.nb_ovn = ovn_api
    ovn_driver.sb_ovn = ovn_sb_api
    ovn_driver._post_fork_event.set()

    synchronizer = ovn_db_sync.OvnNbSynchronizer(
        core_plugin, ovn_api, ovn_sb_api, mode, ovn_driver)

    LOG.info('Neutron OVN Northbound DB sync started with mode: %s', mode)
    synchronizer.do_sync()
    LOG.info('Neutron OVN Northbound DB sync completed')

    sb_synchronizer = ovn_db_sync.OvnSbSynchronizer(
        core_plugin, ovn_sb_api, ovn_driver)

    LOG.info('Neutron OVN Southbound DB sync started with mode: %s', mode)
    sb_synchronizer.do_sync()
    LOG.info('Neutron OVN Southbound DB sync completed')

    if migrate:
        LOG.info("Neutron database migration from OVS to OVN started")
        db_migration.migrate_neutron_database_to_ovn()
        LOG.info("Neutron database migration from OVS to OVN completed")

    LOG.info('Neutron OVN DB sync completed')
